//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// ALU
//
`include "opcodes_def.v"
`include "parms.v"

module alu (
  clk_i, rst_i, rs_i, rt_i, rsa_i, rta_i, rda_i, pc_i, pc_next_i, op_i, sa_i,
  funct_i, offset_i, index_i, reg_hi_i, reg_lo_i, sys_reg_i, kernel_mode_i,
  rd_o, rda_o, pc_o, ioverflow_o, branch_o, va_ex_o, load_o, store_o, rtype_o, va_o,
  reg_hi_o, reg_lo_o, rd_sc_o, reg_hwo_o, reg_lwo_o, sys_reg_o, set_load_o, set_store_o,
  access_ex_o
  );
  input clk_i;              // Input clock
  input rst_i;              // Input reset signal
  input [31:0] rs_i;        // Input source register
  input [31:0] rt_i;        // Input source register
  input [4:0] rsa_i;        // Input address of rs
  input [4:0] rta_i;        // Input address of rt
  input [4:0] rda_i;        // Input address of rd
  input [31:0] pc_i;        // Input Program Counter
  input [31:0] pc_next_i;   // Input Program Counter of the instruction in delay slot
  input [5:0] op_i;         // Input opcode
  input [4:0] sa_i;         // Shift amount
  input [5:0] funct_i;      // Function
  input [15:0] offset_i;    // Offset for branch/jump instructions
  input [25:0] index_i;     // Index for jump instructions  
  input [31:0] reg_hi_i;    // Input hi register from the A stage
  input [31:0] reg_lo_i;    // Input lo register from the A stage
  input [31:0] sys_reg_i;   // Input system register from the CP0
  input kernel_mode_i;      // Signal from MMU that program is in kernel mode
  output [31:0] rd_o;       // Output destination register 
  output [4:0] rda_o;       // Address of the destination register
  output [31:0] pc_o;       // Program Counter for branch/jump operations
  output ioverflow_o;       // Flag to signal overflow
  output branch_o;          // Flag specifying that branch is required
  output va_ex_o;           // Flag specififying that virtual address error has occured
  output load_o;            // Flag to tell to M stage that load from memory is required
  output store_o;           // Flag to tell to M stage that store to memory is required
  output rtype_o;           // Flag to tell that instruction is of R-type
  output [31:0] va_o;       // Virtual address
  output [31:0] reg_hi_o;   // Output hi register to the A stage
  output [31:0] reg_lo_o;   // Output lo register to the A stage
  output [31:0] rd_sc_o;    // Output data for the M stage for SC operation
  output reg_hwo_o;         // Signal to the A stage to take input hi register from ALU
  output reg_lwo_o;         // Signal to the A stage to take input lo register from ALU
  output [31:0] sys_reg_o;         // Output system register to the CP0 from MTC0 instruction
  output set_load_o;        // To tell to sys CP0 that exception was due to LOAD instruction
  output set_store_o;       // To tell to sys CP0 that exception was due to STORE instruction
  output access_ex_o;       // Signal to CP0 that access violation occured
  
  reg [31:0] rd_o; 
  reg [4:0] rda_o;     
  reg [31:0] pc_o;      
  reg branch_o;
  reg va_ex_o;
  reg load_o;
  reg store_o;
  reg rtype_o;
  reg [31:0] va_o;
  reg [31:0] reg_hi_o;
  reg [31:0] reg_lo_o;
  reg [31:0] rd_sc_o;
  reg set_store_o;
  reg set_load_o;
  
  
  // Internal wires and regs
  
  // Opcodes JAL, BLTZAL, BLTZALL, BGEZAL, BGEZALL indirectly
  // use register r31 as destination 
  
  
  // ALU performs decoding of  instructions
  wire a_add = (op_i == `OP_SPECIAL) && (funct_i == `SP_ADD);
  wire a_addi = (op_i == `OP_ADDI);
  wire a_addiu = (op_i == `OP_ADDIU);
  wire a_addu = (op_i == `OP_SPECIAL) &&  (funct_i == `SP_ADDU);
  wire a_and = (op_i == `OP_SPECIAL) && (funct_i == `SP_AND);
  wire a_andi = (op_i == `OP_ANDI);
  wire a_beq = (op_i == `OP_BEQ);
  wire a_bgez = (op_i == `OP_REGIMM) && (rta_i == `R_BGEZ);
  wire a_bgezal = (op_i == `OP_REGIMM) && (rta_i == `R_BGEZAL);
  wire a_bgtz = (op_i == `OP_BGTZ);
  wire a_blez = (op_i == `OP_BLEZ);
  wire a_bltz = (op_i == `OP_REGIMM) && (rta_i == `R_BLTZ);
  wire a_bltzal = (op_i == `OP_REGIMM) && (rta_i == `R_BLTZAL);
  wire a_bne = (op_i == `OP_BNE);
  wire a_clo = (op_i == `OP_SPECIAL2) && (funct_i == `SP2_CLO);
  wire a_clz = (op_i == `OP_SPECIAL2) && (funct_i == `SP2_CLZ);
  wire a_j = (op_i == `OP_J);
  wire a_jal = (op_i == `OP_JAL);
  wire a_jalr = (op_i == `OP_SPECIAL) && (funct_i == `SP_JALR);
  wire a_jr = (op_i == `OP_SPECIAL) && (funct_i == `SP_JR);
  wire a_lb = (op_i == `OP_LB);
  wire a_lbu = (op_i == `OP_LBU);
  wire a_lh = (op_i == `OP_LH);
  wire a_lhu = (op_i == `OP_LHU);
  wire a_ll = (op_i == `OP_LL);
  wire a_lui = (op_i == `OP_LUI);
  wire a_lw = (op_i == `OP_LW);
  wire a_lwl = (op_i == `OP_LWL);
  wire a_lwr = (op_i == `OP_LWR);
  wire a_mfc0 = (op_i == `OP_COP0) && (rsa_i == `CZ_MFC0);
  wire a_mfhi = (op_i == `OP_SPECIAL) && (funct_i == `SP_MFHI);
  wire a_mflo = (op_i == `OP_SPECIAL) && (funct_i == `SP_MFLO);
  wire a_movn = (op_i == `OP_SPECIAL) && (funct_i == `SP_MOVN);
  wire a_movz = (op_i == `OP_SPECIAL) && (funct_i == `SP_MOVZ);
  wire a_mtc0 = (op_i == `OP_COP0) && (rsa_i == `CZ_MTC0); 
  wire a_mthi = (op_i == `OP_SPECIAL) && (funct_i == `SP_MTHI);
  wire a_mtlo = (op_i == `OP_SPECIAL) && (funct_i == `SP_MTLO);
  wire a_nor = (op_i == `OP_SPECIAL) && (funct_i == `SP_NOR);
  wire a_or = (op_i == `OP_SPECIAL) && (funct_i == `SP_OR);
  wire a_ori = (op_i == `OP_ORI);
  wire a_sb = (op_i == `OP_SB);
  wire a_sc = (op_i == `OP_SC);
  wire a_sh = (op_i == `OP_SH);
  wire a_sll = (op_i == `OP_SPECIAL) && (funct_i == `SP_SLL);
  wire a_sllv = (op_i == `OP_SPECIAL) && (funct_i == `SP_SLLV);
  wire a_slt = (op_i == `OP_SPECIAL) && (funct_i == `SP_SLT);
  wire a_slti = (op_i == `OP_SLTI);
  wire a_sltiu = (op_i == `OP_SLTIU);
  wire a_sltu = (op_i == `OP_SPECIAL) && (funct_i == `SP_SLTU);
  wire a_sra = (op_i == `OP_SPECIAL) && (funct_i == `SP_SRA);
  wire a_srav = (op_i == `OP_SPECIAL) && (funct_i == `SP_SRAV);
  wire a_srl = (op_i == `OP_SPECIAL) && (funct_i == `SP_SRL);
  wire a_srlv = (op_i == `OP_SPECIAL) && (funct_i == `SP_SRLV);
  wire a_sub = (op_i == `OP_SPECIAL) && (funct_i == `SP_SUB);
  wire a_subu = (op_i == `OP_SPECIAL) && (funct_i == `SP_SUBU);
  wire a_sw = (op_i == `OP_SW);
  wire a_swl = (op_i == `OP_SWL);
  wire a_swr = (op_i == `OP_SWR);
  wire a_xor = (op_i == `OP_SPECIAL) && (funct_i == `SP_XOR);
  wire a_xori = (op_i == `OP_XORI);
  
  wire [31:0] sign_exts_offset = {{16{offset_i[15]}},offset_i[15:0]};   // Sign extended and shifted 2 bits left offset
  
  wire set_branch = a_beq || a_jal || a_bgez      // This wire OR'es all the commands which cause branch. If non of them is active no branch_o must be set to zero
                    || a_bgezal || a_bgtz || a_blez || a_bltz || a_bltzal || a_bne || a_j
                    || a_jalr || a_jr;
  wire set_load = a_lw || a_lwl || a_lwr || a_lb || a_lbu || a_lh || a_lhu || a_ll;
  wire set_store = a_sw || a_sb || a_sh || a_swl || a_swr || a_sc;
  // Instructions which cause regfile to be updated. Destination register rd or rt (rtype or itype instructions)
  wire set_rtype = a_add || a_addu || a_addi || a_addiu || a_jal || a_clo || a_clz || a_lui 
                    || a_sub || a_subu || a_sll || a_sllv || a_sra || a_srav || a_srl || a_srlv
                    || a_and || a_andi || a_nor || a_or || a_ori || a_xor || a_xori || a_movn
                    || a_movz || a_slt || a_slti || a_sltiu || a_sltu || a_mfhi || a_mflo || a_bgezal
                    || a_bltzal || a_jalr || a_sc || a_mfc0;
  wire oflow_instr = a_add || a_addi || a_sub;                 // Wire to join all instructions which may cause overflow
  wire dest_rt = a_addi || a_addiu || a_lui || a_andi || a_ori                    // Registers which have rt as destination instead of rd
                 || a_xori || a_slti || a_sltiu || a_sc || a_mfc0;
  
  reg [31:0] rsi_buf;                 // 0.5 clk delay
  reg [31:0] rti_buf;                 // 0.5 clk delay
  reg oflow_instr_buf;                // 0.5 clk delay
  reg dest_rt_buf;                    // 0.5 clk delay
  reg [4:0] rta_buf;                  // 0.5 clk delay
  reg [4:0] rda_buf;                  // 0.5 clk delay
  
  reg carry_out;
  wire carry_in = ((rsi_buf[31]^rti_buf[31])^rd_o[31]);
  assign ioverflow_o = oflow_instr_buf & (carry_out^carry_in);     // Overflow detect for 2's complement logic
                // But only for those instructions which cause overflow
  assign access_ex_o = (va_o[31] & ~kernel_mode_i) && (load_o | store_o);   // Access violation for data accesses
                
  wire [31:0] clzo_buf = (a_clo) ? rs_i: ~rs_i;       // In order to use the same case condition for both CLO and CLZ
  
  assign reg_hwo_o = a_mthi;
  assign reg_lwo_o = a_mtlo;
  
  assign sys_reg_o = (a_mtc0) ? rt_i: 32'hzzzzzzzz;
                
  always @ (*)        // Logic to specify rda output
    begin
      if (dest_rt_buf)
        rda_o = rta_buf;
      else
        rda_o = rda_buf;
    end
  
  always @ (posedge clk_i)
    begin
      case (1'b1) // synthesis full_case
        a_add, a_addu: {carry_out,rd_o} <=  {1'b0,rs_i} + {1'b0,rt_i};
        a_addi, a_addiu: {carry_out,rd_o} <= {1'b0,rs_i} + {1'b0,sign_exts_offset};
        a_and: rd_o <= rs_i & rt_i;
        a_andi: rd_o <= rs_i & {16'h0,offset_i};
        a_beq, a_bgez, a_bgezal, a_bgtz, a_blez, a_bltz, a_bltzal, a_bne:                          
          begin
            if (((rs_i == rt_i) && a_beq) || ((rs_i[31] == 1'b0) && (a_bgez|a_bgezal)) ||
                  ((rs_i[31] == 1'b0) && (rs_i[30:0] != 31'b0) && a_bgtz) ||
                  (((rs_i[31] == 1'b1) || (rs_i[30:0] == 31'b0)) && a_blez)  ||
                  ((rs_i[31] == 1'b1) && (a_bltz|a_bltzal)) || ((rs_i != rt_i) && a_bne))
              begin
                pc_o <= pc_next_i + {sign_exts_offset[29:0],2'b00};  
                branch_o <= 1'b1;
                rd_o <= pc_i + 4'h8;                // Meaningful for BGEZAL
              end
          end 
        a_clo, a_clz:
          begin
            casex (clzo_buf)          // Synthesis full case
             `B32:     rd_o <= 32'h20;
             `B31:     rd_o <= 32'h1F;
             `B30:     rd_o <= 32'h1E;
             `B29:     rd_o <= 32'h1D;
             `B28:     rd_o <= 32'h1C;
             `B27:     rd_o <= 32'h1B;
             `B26:     rd_o <= 32'h1A;
             `B25:     rd_o <= 32'h19;
             `B24:     rd_o <= 32'h18;
             `B23:     rd_o <= 32'h17;
             `B22:     rd_o <= 32'h16;
             `B21:     rd_o <= 32'h15;
             `B20:     rd_o <= 32'h14;
             `B19:     rd_o <= 32'h13;
             `B18:     rd_o <= 32'h12;
             `B17:     rd_o <= 32'h11;
             `B16:     rd_o <= 32'h10;
             `B15:     rd_o <= 32'hF;
             `B14:     rd_o <= 32'hE;
             `B13:     rd_o <= 32'hD;
             `B12:     rd_o <= 32'hC;
             `B11:     rd_o <= 32'hB;
             `B10:     rd_o <= 32'hA;
             `B9:      rd_o <= 32'h9;
             `B8:      rd_o <= 32'h8;
             `B7:      rd_o <= 32'h7;
             `B6:      rd_o <= 32'h6;
             `B5:      rd_o <= 32'h5;
             `B4:      rd_o <= 32'h4;
             `B3:      rd_o <= 32'h3;
             `B2:      rd_o <= 32'h2;
             `B1:      rd_o <= 32'h1;
             default: rd_o <= 32'h0;
            endcase  
          end   
        a_jal, a_j:                      
          begin
            rd_o <= pc_i + 4'h8;
            pc_o <= {pc_next_i[31:28],index_i[25:0],2'b00};
            branch_o <= 1'b1; 
          end
        a_jalr, a_jr:
          begin
            rd_o <= pc_i + 4'h8;
            pc_o <= rs_i;
            branch_o <= 1'b1;
          end
        a_lui: rd_o <= {offset_i[15:0],16'b0};
        a_lw, a_lh, a_lhu, a_sh, a_sw, a_ll, a_sc:
          begin
            if  ((({offset_i[1:0],rs_i[1:0]} != 4'b0000) && (a_lw|a_sw|a_ll|a_sc)) || 
                    (({offset_i[0],rs_i[0]} != 2'b00) && (a_lh|a_lhu|a_sh)))
              begin
                va_ex_o <= 1'b1;    // Wrong offset has been provided
                load_o <= 1'b0;
                store_o <= 1'b0;
                va_o <= rs_i + sign_exts_offset;      // Bad VA is required by CP0
              end
            else
              begin
                va_o <= rs_i + sign_exts_offset;
                va_ex_o <= 1'b0;
                if (a_lw|a_lh|a_lhu|a_ll)
                  load_o <= 1'b1;
                else
                  begin
                    store_o <= 1'b1;
                    if (a_sc)
                      begin
                        rd_o <= 32'b1;
                        rd_sc_o <= rt_i;
                      end
                    else
                      rd_o <= rt_i;
                  end
              end
          end
        a_lwl, a_lwr, a_lb, a_lbu, a_sb, a_swl, a_swr:
          begin
            va_o <= rs_i + sign_exts_offset;
            rd_o <= rt_i;                       // Will be required in  alignment procedure
            va_ex_o <= 1'b0;                    // These instructions never cause alignment error
            if (a_sb|a_swl|a_swr)
              store_o <= 1'b1;
            else
              load_o <= 1'b1;
          end
        a_mfc0: rd_o <= sys_reg_i; 
        a_mfhi: rd_o <= reg_hi_i;
        a_mflo: rd_o <= reg_lo_i;
        a_mthi: reg_hi_o <= rs_i;
        a_mtlo: reg_lo_o <= rs_i;
        a_nor: rd_o <= rs_i | (~rt_i);
        a_or: rd_o <= rs_i | rt_i;
        a_ori: rd_o <= rs_i | {16'h0,offset_i};
        a_sll: rd_o <= rt_i << sa_i;
        a_sllv: rd_o <= rt_i << rs_i[4:0];
        a_slt:
          begin
            if (rs_i[31] > rt_i[31])          // Signed compare
              rd_o <= 32'h1;
            else if (rs_i[31] < rt_i[31])
              rd_o <= 32'h0;
            else if (rs_i[30:0] < rt_i[30:0])
              rd_o <= 32'h1;
            else
              rd_o <= 32'h0;
          end
        a_slti:
          begin
            if (rs_i[31] > sign_exts_offset[31])  // Signed compare
              rd_o <= 32'h1;
            else if (rs_i[31] < sign_exts_offset[31])
              rd_o <= 32'h0;
            else if (rs_i[30:0] < sign_exts_offset[30:0])
              rd_o <= 32'h1;
            else 
              rd_o <=32'h0;
          end
        a_sltiu:
          begin
            if (rs_i < sign_exts_offset)
              rd_o <= 32'h1;
            else
              rd_o <= 32'h0;
          end
        a_sltu:
          begin
            if (rs_i < rt_i)
              rd_o <= 32'h1;
            else
              rd_o <= 32'h0;
          end
        a_sra: rd_o <= {{32{rt_i[31]}},rt_i} >> sa_i ;    // Assign 32 bit LSB of 64bit vector
        a_srav: rd_o <= {{32{rt_i[31]}},rt_i} >> rs_i[4:0];   // Assign 32 bit LSB of 64bit vector
        a_srl:  rd_o <= rt_i >> sa_i;
        a_srlv: rd_o <= rt_i >> rs_i[4:0];
        a_sub, a_subu: {carry_out,rd_o} <= {1'b0,rs_i} - {1'b0,rt_i};
        a_xor: rd_o <= rs_i ^ rt_i;
        a_xori: rd_o <= rs_i ^ {16'h0,offset_i};
          
      endcase
      
      if (!set_branch)
        branch_o <= 1'b0;
        
      if (!set_load)
        load_o <= 1'b0;
        
      if (!(set_load|set_store))
        va_ex_o <= 1'b0;
        
      if (!set_store)
        store_o <= 1'b0;
        
      if (set_rtype & rst_i)    // rst_i here is required, otherwise rtype_o will be set to 1 on reset
                                // as NOP (SLL) is an instruction of rtype.
        case (1'b1)             // Synthesis full case
          a_movn:
            begin
              if (rt_i !== 32'h0)
                begin
                  rd_o <= rs_i;
                  rtype_o <= 1'b1;
                end
              else
                rtype_o <= 1'b0;          // Don't update rd
            end
          a_movz:
            begin
              if (rt_i === 32'h0)
                begin
                  rd_o <= rs_i;
                  rtype_o <= 1'b1;
                end
              else
                rtype_o <= 1'b0;
            end
          a_bgezal:
            begin
              if (rs_i[31] === 1'b0)
                rtype_o <= 1'b1;
              else
                rtype_o <= 1'b0;
            end
          a_bltzal:
            begin
              if (rs_i[31] === 1'b1)
                rtype_o <= 1'b1;
              else
                rtype_o <= 1'b0;
            end
          default: rtype_o <= 1'b1;
        endcase
      else
        rtype_o <= 1'b0;
        
        
      rsi_buf <= rs_i;
      
      if (a_addi)
        rti_buf <= sign_exts_offset;
      else
        rti_buf <= rt_i;
      
      rta_buf <= rta_i;                           // 0.5clk delay      
      rda_buf <= rda_i;                           // 0.5clk delay
      oflow_instr_buf <= oflow_instr;
      dest_rt_buf <= dest_rt;
       
      set_load_o <= set_load;
      set_store_o <= set_store;  
    end
  
endmodule
