//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Align stage of the MIPS32-4kp execution core
//
`include "opcodes_def.v"

module a_stage (
  clk_i, rst_i, rd_i, rda_i, rtype_i, load_i, 
  opcode_i, rdl_i, rdla_i, offset01_i, reg_hi_i, reg_lo_i,
  reg_hia_i, reg_loa_i, reg_hwa_i, reg_lwa_i, reg_hwm_i, reg_lwm_i, mul_i, rdamul_i,
  rd_o, rda_o, write_o, rdl_aligned_o, reg_hi_o, reg_lo_o
  );
  input clk_i;              // Input clock
  input rst_i;              // Input reset signal
  input [31:0] rd_i;        // Input RD register
  input [4:0] rda_i;        // Input address of the RD register
  input rtype_i;            // Flag that instruction is of R-type
  input load_i;             // Flag that instruction is Load
  input [5:0] opcode_i;     // Input opcode from the M stage
  input [31:0] rdl_i;       // Input after load operation
  input [4:0] rdla_i;       // Input address for load register
  input [1:0] offset01_i;   // Two LSB of register's virtual address for Load operation
  input [31:0] reg_hi_i;    // Input hi register from MDU
  input [31:0] reg_lo_i;    // Input lo register from MDU
  input [31:0] reg_hia_i;   // Input hi register from ALU
  input [31:0] reg_loa_i;   // Input lo register from ALU
  input reg_hwa_i;          // Signal from ALU to load hi register
  input reg_lwa_i;          // Signal from ALU to load lo register
  input reg_hwm_i;          // Signal from MDU to load hi register
  input reg_lwm_i;          // Signal from MDU to load lo register 
  input mul_i;              // MUL instruction in MDU
  input [4:0] rdamul_i;     // rd from MUL instruction
  output [31:0] rd_o;       // Output RD register         
  output [4:0] rda_o;       // Output address of the RD register
  output write_o;           // Signal to write data to the W stage
  output rdl_aligned_o;     // Aligned output for the bypass module and for W stage as well
  output [31:0] reg_hi_o;   // Output hi register to the MDU
  output [31:0] reg_lo_o;   // Output lo register to the MDU
  
  reg [31:0] rd_o;
  reg [4:0] rda_o;
  reg write_o;
  reg [31:0] reg_hi_o;
  reg [31:0] reg_lo_o;
  
  // Internal wires and registers
  reg [31:0] rd_buf;                 // 0.5 clk delay
  reg [4:0] rda_buf;                 // 0.5 clk delay
  reg rtype_buf;                       // 0.5 clk delay
  reg load_buf;                        // 0.5 clk delay
  
  reg [31:0] rdl_aligned_o;           
  
  always @ (opcode_i or offset01_i)                      // Data aligner
    begin                   // If (load_i) statement is not required as opcode_i
                            // already specifies load operation
      case (opcode_i)               // Synthesis full case
        `OP_LB:
          case (offset01_i)
            2'b00: rdl_aligned_o = {{24{rdl_i[7]}},rdl_i[7:0]};
            2'b01: rdl_aligned_o = {{24{rdl_i[15]}},rdl_i[15:8]};
            2'b10: rdl_aligned_o = {{24{rdl_i[23]}},rdl_i[23:16]};
            2'b11: rdl_aligned_o = {{24{rdl_i[31]}},rdl_i[31:24]};
          endcase
        `OP_LBU:
          case (offset01_i)
            2'b00: rdl_aligned_o = {24'b0,rdl_i[7:0]};
            2'b01: rdl_aligned_o = {24'b0,rdl_i[15:8]};
            2'b10: rdl_aligned_o = {24'b0,rdl_i[23:16]};
            2'b11: rdl_aligned_o = {24'b0,rdl_i[31:24]};
          endcase
        `OP_LH:
          case (offset01_i)       // Synthesis full case
            2'b00: rdl_aligned_o = {{16{rdl_i[15]}},rdl_i[15:0]}; 
            2'b10: rdl_aligned_o = {{16{rdl_i[31]}},rdl_i[31:16]};
          endcase
        `OP_LHU:
          case (offset01_i)       // Synthesis full case
            2'b00: rdl_aligned_o = {16'b0,rdl_i[15:0]};
            2'b10: rdl_aligned_o = {16'b0,rdl_i[31:16]};
          endcase
        `OP_LW, `OP_LL:   rdl_aligned_o = rdl_i;
        `OP_LWL:
          case (offset01_i)
            2'b00:  rdl_aligned_o = {rdl_i[7:0],rd_i[23:0]};
            2'b01:  rdl_aligned_o = {rdl_i[15:0],rd_i[15:0]};
            2'b10:  rdl_aligned_o = {rdl_i[23:0],rd_i[7:0]};
            2'b11:  rdl_aligned_o = rdl_i;
          endcase
        `OP_LWR:
          case (offset01_i)
            2'b00: rdl_aligned_o = rdl_i;
            2'b01: rdl_aligned_o = {rd_i[31:24],rdl_i[31:8]};
            2'b10: rdl_aligned_o = {rd_i[31:16],rdl_i[31:16]};
            2'b11: rdl_aligned_o = {rd_i[31:8],rdl_i[31:24]};
          endcase
        
          
      endcase
    end
  
  always @ (posedge clk_i)
    begin
      if (load_i)   // Need to decide what to send to the W-stage
        begin
          rd_buf <= rdl_aligned_o;      // Aligned load results
          rda_buf <= rdla_i;
        end
      else if (rtype_i)
        begin
          rd_buf <= rd_i;             // R-type operation result
          rda_buf <= rda_i;
        end
      rtype_buf <= rtype_i;
      load_buf <= load_i;

    end
    
  always @ (negedge clk_i)
    begin
      if (mul_i)
        begin
          rd_o <= reg_lo_i;
          rda_o <= rdamul_i;
        end
      else
        begin
          rd_o <= rd_buf;
          rda_o <= rda_buf;
        end

      write_o <= (load_i | rtype_i | mul_i);
      
      if (!rst_i)
        begin
          reg_hi_o <= 32'h0;
          reg_lo_o <= 32'h0;
        end
      else
        begin
          if (reg_hwa_i)
            reg_hi_o <= reg_hia_i;
          else if (reg_hwm_i)
            reg_hi_o <= reg_hi_i;
          
          if (reg_lwa_i)
            reg_lo_o <= reg_loa_i;
          else if (reg_lwm_i)
            reg_lo_o <= reg_lo_i;
        end
    end
      
endmodule
