//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Bus interface unit
//
module biu (
  clk_i, cold_rst_i, dataex_i, datam_i, pham_i, rdi1_i, rdi4_i, rdd1_i,
  eb_ardy_i, eb_rdval_i, eb_rberr_i, wrd1_i, byte_mask_i, eb_wdrdy_i, eb_wberr_i,
  uart_txrdy_i,
  dataex_o, adrex_o, datam_o, rdi1_ready_o, rdi4_ready_o, eb_instr_o, eb_be_o,
  eb_avalid_o, eb_write_o, eb_rberr_o, eb_burst_o, eb_bfirst_o, eb_blast_o,
  rdd1_ready_o, wrd1_done_o, eb_wberr_o, uart_wr_o, uart_data_o
  );
  
  input clk_i;                      // Input clock
  input cold_rst_i;                 // Input cold reset signal
  input [31:0] dataex_i;            // Input data from external source
  input [31:0] datam_i;             // Input data from MMU
  input [31:0] pham_i;              // Input physical address from MMU
  input rdi1_i;                     // Request from MMU to fetch one instruction
  input rdi4_i;                     // Request from MMU to fetch four instructions
  input rdd1_i;                     // Request from MMU to read one data word
  input eb_ardy_i;                  // Indicates whether target is ready for new address 
  input eb_rdval_i;                 // Indicates that target is driving data 
  input eb_rberr_i;                 // Bus error indicator for read transaction
  input wrd1_i;                     // Request to write one word to memory
  input [3:0] byte_mask_i;          // Byte mask input from MMU
  input eb_wdrdy_i;                 // Input signal from external memory that write cycle is being done
  input eb_wberr_i;                 // Signal from external memory that bus error occurred during write cycle
  input uart_txrdy_i;               // Input from UART
  output [31:0] dataex_o;           // Output data to external memory
  output [31:0] adrex_o;            // Output address
  output [31:0] datam_o;            // Output data for MMU
  output rdi1_ready_o;              // Output for MMU to signal that one instruction is fetched
  output rdi4_ready_o;              // Output for MMU to signal that four instruction are being read
  output eb_instr_o;                // When asserted indicates that data read is an instruction fetch
  output [3:0] eb_be_o;             // Byte enable
  output eb_avalid_o;               // When asserted indicates start of new bus transaction
  output eb_write_o;                // When asserted, indicates that current transaction is write
  output eb_rberr_o;                // Signal to CP0 that bus error during read cycle occured
  output eb_burst_o;                // Current transaction is part of a cache fill or a write burst
  output eb_bfirst_o;               // Beginning of the burst
  output eb_blast_o;                // End of the burst
  output rdd1_ready_o;              // Signal for MMU that one data word is read
  output wrd1_done_o;               // Signal for MMU that write cycle is completed
  output eb_wberr_o;                // Signal that external memory write error occured
  output uart_wr_o;                 // Output to UART
  output [7:0] uart_data_o;         // Output data to UART
  
  reg rdi1_ready_o;
  reg rdi4_ready_o;
  reg eb_instr_o;
  reg [3:0] eb_be_o;
  reg eb_avalid_o;
  reg eb_write_o;
  reg [31:0] datam_o;
  reg eb_rberr_o;
  reg eb_burst_o;
  reg eb_bfirst_o;
  reg eb_blast_o;
//  reg [31:0] adrex_o;
  reg [31:0] dataex_o;
  reg rdd1_ready_o;
  reg wrd1_done_o;
  reg eb_wberr_o;
  reg rdi1_ibuf;                    // To shift transition of rdi1_i to the negedge
  reg rdi4_ibuf;                    // To shift transition of rdi4_i to the negedge
  reg rdd1_ibuf;                    // To shift transition of rdd1_i to the negedge
  reg wrd1_ibuf;                    // To shift transition of wrd1_i to the negedge
  reg eb_ardyi_pos;                 // To shift transition of eb_ardy_i to the posedge
  reg [7:0] bus_count;              // Counter for bus transactions
  reg [1:0] word_align;             // Word align for burst transaction, to count successive words
  reg [1:0] word_align_neg;         // to shift word_align transition to the negedge
  reg [31:0] phami_buf;             // To record value of physical address at the beginning of the burst
  reg eb_wdrdyi_pos;                // Buffer to shift transition of eb_wdrdy_i to the posedge
  reg eb_wdrdyi_neg;                // To delay eb_wdrdy_i to one clock in order to align eb_wdrdy_i and 
                                    // eb_wberr_i 
  
  assign adrex_o = (rdi4_i) ? {pham_i[31:4],word_align[1:0],pham_i[1:0]} : pham_i;

  // Test logic ---------------------------------------
  reg [3:0] sw_r = 4'h0;
  reg [3:0] strap_r = 4'h0;
  reg [9:0] count_r = 10'h41;
  reg [31:0] datami_buf;
  
    test_m uart_link (
    .clk_i                (clk_i),
    .rst_i                (cold_rst_i),
    .sdata1_i             (dataex_o),
    .sdata2_i             (datami_buf),
    .sw_i                 (sw_r),
    .strap_i              (strap_r),
    .count_i              (count_r),
    .uart_txrdy_i         (uart_txrdy_i),
    .start_i              (cold_rst_i),
    .uart_wr_o            (uart_wr_o),
    .uart_data_o          (uart_data_o)
  );                    
  
  always @ (posedge clk_i)
    begin
      datami_buf <= datam_i;
    end
 //-------------------------------------------------------
 
 
  always @ (posedge clk_i)
    begin
      if (!cold_rst_i)
        begin
          rdi1_ready_o <= 1'b0;
          rdi4_ready_o <= 1'b0;
          rdd1_ready_o <= 1'b0;
          eb_rberr_o <= 1'b0;
          eb_bfirst_o <= 1'b0;
          eb_blast_o <= 1'b0;
          eb_avalid_o <= 1'b0;
          wrd1_done_o <= 1'b0;
          word_align <= 2'b00;
          eb_wberr_o <= 1'b0;
        end
      else
        begin
          case (1'b1)             // Synthesis full case
            rdi1_ibuf:            // Request to fetch one instruction
              begin
                eb_instr_o <= 1'b1;
                eb_be_o <= 4'b1111;
                eb_write_o <= 1'b0;
                eb_burst_o <= 1'b0;
                if (bus_count > 8'h01) 
                  eb_avalid_o <= 1'b0;
                else
                  eb_avalid_o <= 1'b1;
                if(eb_rdval_i & ~eb_rberr_i)
                  begin
                    datam_o <= dataex_i;
                    rdi1_ready_o <= 1'b1;
                  end
                if (eb_rberr_i)
                  eb_rberr_o <= 1'b1;         // Error occurred
              end
            rdi4_ibuf:          // Request to refill icache
              begin
                case ({bus_count,eb_ardy_i})          // Synthesis full case
                  9'b000000011:
                    begin
                      word_align <= pham_i[3:2];
                      phami_buf <= pham_i;
                      eb_burst_o <= 1'b1;
                      eb_bfirst_o <= 1'b1;
                      eb_blast_o <= 1'b0;
                      eb_write_o <= 1'b0;
                      eb_be_o <= 4'b1111;
                      eb_instr_o <= 1'b1;
                    end
                  9'b000000101, 9'b000000111, 9'b000001001:
                    begin
                      word_align <= word_align_neg;
                      eb_bfirst_o <= 1'b0;
                      if (bus_count == 8'h04)
                        eb_blast_o <= 1'b1;
                      else
                        eb_blast_o <= 1'b0;
                    end 
                  default:
                    begin
                      eb_blast_o <= 1'b0;
                    end 
                endcase
                if (bus_count > 8'h04)
                  eb_avalid_o <= 1'b0;
                else
                  eb_avalid_o <= 1'b1;
                               
                if (eb_rdval_i & ~eb_rberr_i)
                  begin
                    rdi4_ready_o <= 1'b1;
                    datam_o <= dataex_i;
                  end
                if (eb_rberr_i)
                  eb_rberr_o <= 1'b1;         // Error occurred
              end
            rdd1_ibuf:
              begin
                eb_instr_o <= 1'b0;
                eb_be_o <= 4'b1111;
                eb_write_o <= 1'b0;
                eb_burst_o <= 1'b0;
                rdi1_ready_o <= 1'b0;       // As rdd1_i may come straight after rdi these
                rdi4_ready_o <= 1'b0;       // lines are required to substitute default deasserts
                if (bus_count > 8'h01)      // One clock less than for rdi1_ibuf
                  eb_avalid_o <= 1'b0;
                else
                  eb_avalid_o <= 1'b1;
                if (eb_rdval_i & ~eb_rberr_i)
                  begin
                    datam_o <= dataex_i;
                    rdd1_ready_o <= 1'b1;
                  end
                if (eb_rberr_i)
                  eb_rberr_o <= 1'b1;       // Bus error occurred
                else
                  eb_rberr_o <= 1'b0;
              end
            wrd1_ibuf:
              begin
                eb_be_o <= byte_mask_i;
                eb_write_o <= 1'b1;
                eb_burst_o <= 1'b0;
                wrd1_done_o <= 1'b0;
                dataex_o <= datam_i;
                rdi1_ready_o <= 1'b0;       // As wrd1_i may come straight after rdi these
                rdi4_ready_o <= 1'b0;       // lines are required to substitute default deasserts
                if (bus_count > 8'h01)
                  eb_avalid_o <= 1'b0;
                else 
                  eb_avalid_o <= 1'b1;
                if (eb_wdrdyi_neg & ~eb_wberr_i)
                  wrd1_done_o <= 1'b1;
                else
                  wrd1_done_o <= 1'b0;
                if (eb_wberr_i)
                  eb_wberr_o <= 1'b1;
                else
                  eb_wberr_o <= 1'b0;
                
              end
            default:
              begin
                eb_instr_o <= 1'b0;
                eb_be_o <= 4'b0000;
                eb_write_o <= 1'b0;
                eb_avalid_o <= 1'b0;
                rdi1_ready_o <= 1'b0;
                eb_rberr_o <= 1'b0;
                rdi4_ready_o <= 1'b0;
                eb_burst_o <= 1'b0;
                rdd1_ready_o <= 1'b0;
                wrd1_done_o <= 1'b0;
              end
          endcase
              
        end
      eb_ardyi_pos <= eb_ardy_i;
      eb_wdrdyi_pos <= eb_wdrdy_i;
    end
  
  always @ (negedge clk_i)
    begin
      if (!cold_rst_i)
        begin
          bus_count <= 8'h00;
        end
      else
        begin
         if ((eb_ardyi_pos & rdi1_i) || (eb_ardyi_pos & rdi4_i) || (eb_ardyi_pos & rdd1_i) ||
             (eb_ardyi_pos & wrd1_i)) 
           bus_count <= bus_count + 1'b1;
         else if (!rdi4_i)
           bus_count <= 8'h00;
         word_align_neg <= word_align + 1'b1;  
        end
        
      rdi1_ibuf <= rdi1_i;
      rdi4_ibuf <= rdi4_i;
      rdd1_ibuf <= rdd1_i;
      wrd1_ibuf <= wrd1_i;
      eb_wdrdyi_neg <= eb_wdrdyi_pos;
    end
    
  
endmodule