//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// E Stage of the MIPS32-4kp execution core
// Stage starts on positive edge of clock
//
`include "opcodes_def.v"
//
module e_stage (
  clk_i, rst_i, opcode_i, rdw_i, rdaw_i, wen_i, rdms_i, rdas_i,
  pc_i, pc_next_i, rs_bps_i, rt_bps_i, reg_hi_i, reg_lo_i, stall_i, slip_i,
  sys_reg_i, kernel_mode_i, uart_txrdy_i,
  rd_o, rda_o, pc_o, ioverflow_o, opcode_o, branch_o, va_ex_o, load_o,
  store_o, rtype_o, va_o, rs_reg_o, rt_reg_o, reg_hi_o, reg_lo_o, rd_sc_o, 
  reg_hwo_o, reg_lwo_o, sys_reg_o, set_load_o, set_store_o, set_branch_o,
  access_ex_o, uart_wr_o, uart_data_o
  );
  input clk_i;    // Input clock
  input rst_i;    // Reset signal
  input [31:0] opcode_i;    // Opcode from the istage
  input [31:0] rdw_i;       // Input rd register from W stage
  input [4:0] rdaw_i;       // Address for the destination register taken from W stage
  input wen_i;              // Write enable singal for the register file from W stage
  input [31:0] rdms_i;      // Input rd register from M stage
  input [31:0] rdas_i;      // Input rd register from A stage
  input [31:0] pc_i;        // Program counter input
  input [31:0] pc_next_i;   // Program counter of the instruction in delay slot  
  input [31:0] rs_bps_i;    // rs from the bypass module
  input [31:0] rt_bps_i;    // rt from the bypass module
  input [31:0] reg_hi_i;    // Input hi register from the A stage
  input [31:0] reg_lo_i;    // Input lo register from the A stage
  input stall_i;            // Input to stall pipeline
  input slip_i;             // Input to slip pipeline
  input [31:0] sys_reg_i;   // Input system register from CP0
  input kernel_mode_i;      // Signal from MMU that program is in kernel mode
  input uart_txrdy_i;       // Input from UART
  output [31:0] rd_o;       // Output RD register from ALU
  output [4:0] rda_o;       // Output address of the RD register
  output [31:0] pc_o;       // Modified PC to be used for the next PC calculation
  output ioverflow_o;       // Overflow from ALU
  output [31:0] opcode_o;   // Opcode to the M stage
  output branch_o;          // Signals that branch condition met
  output va_ex_o;           // Virtual address exception
  output load_o;            // Signal to the M stage that load is required
  output store_o;           // Signal to the M stage that store is required
  output rtype_o;           // Intruction was of R-type
  output [31:0] va_o;       // Virtual address to the M stage
  output [31:0] rs_reg_o;   // rs from the register file
  output [31:0] rt_reg_o;   // rt from the register file
  output [31:0] reg_hi_o;   // Output hi register to the A stage
  output [31:0] reg_lo_o;   // Output lo register to the A stage
  output [31:0] rd_sc_o;    // Output data to be stored in the M stage from SC opcode
  output reg_hwo_o;         // Signal to the A stage to take signal from ALU
  output reg_lwo_o;         // Signal to the A stage to take signal from ALU  
  output [31:0] sys_reg_o;  // Output system register to CP0 from MTC0 instruction
  output set_load_o;        // To tell to the sys CP0 that exception due to the LOAD instruction
  output set_store_o;       // To tell to the sys CP0 that exception due to the STORE instruction
  output set_branch_o;      // Flag to the sys CP0 that instruction is in delay slot
  output access_ex_o;       // Signal to CP0 that access violation occured 
  output uart_wr_o;         // Output to UART
  output [7:0] uart_data_o; // Output to UART
  
  reg [31:0] opcode_o;
  reg [31:0] va_o;
  reg load_o;
  reg store_o;
  reg rtype_o;
  reg [31:0] rd_o;
  reg [31:0] rd_sc_o;
  reg [4:0] rda_o;
  reg set_branch_o;
  
  // Internal wires and regs
  
  wire [31:0] va_buf;         // To add 0.5 clk delay. Register declared in ALU
  wire load_buf;              // Delay for the M stage
  wire store_buf;             // Delay for the M stage
  wire rtype_buf;             // Delay for 0.5 clk
  reg [31:0] opcode_buf;      // To add 0.5 clk delay
  wire [31:0] rd_buf;          // To add 0.5 clk delay
  wire [31:0] rd_sc_buf;      // To add 0.5 clk delay
  wire [4:0] rda_buf;         // To add 0.5 clk delay
  
  // Instantiate modules
  regfile e_regfile(      // Register file
    .clk_i      (clk_i),			  
    .rst_i      (rst_i),
    .rd_i       (rdw_i), 
    .rda_i      (rdaw_i), 
    .rsa_i      (opcode_i[25:21]), 
    .rta_i      (opcode_i[20:16]), 
    .wen_i      (wen_i),
    .rdms_i     (rdms_i), 
    .rdas_i     (rdas_i), 
    .uart_txrdy_i (uart_txrdy_i),
    .rs_o       (rs_reg_o),
    .rt_o       (rt_reg_o),
    .uart_wr_o  (uart_wr_o),
    .uart_data_o (uart_data_o)
  );
  
  alu e_alu (       // ALU
    .clk_i      (clk_i), 
    .rst_i      (rst_i),
    .rs_i       (rs_bps_i), 
    .rt_i       (rt_bps_i),
    .rsa_i      (opcode_i[25:21]),
    .rta_i      (opcode_i[20:16]),
    .rda_i      (opcode_i[15:11]), 
    .pc_i       (pc_i),
    .pc_next_i  (pc_next_i), 
    .op_i       (opcode_i[31:26]), 
    .sa_i       (opcode_i[10:6]),  
    .funct_i    (opcode_i[5:0]), 
    .offset_i   (opcode_i[15:0]),
    .index_i    (opcode_i[25:0]), 
    .reg_hi_i   (reg_hi_i),
    .reg_lo_i   (reg_lo_i),
    .sys_reg_i  (sys_reg_i), 
    .kernel_mode_i  (kernel_mode_i), 
    .rd_o       (rd_buf), 
    .rda_o      (rda_buf),
    .pc_o       (pc_o), 
    .ioverflow_o (ioverflow_o),
    .branch_o   (branch_o),
    .va_ex_o    (va_ex_o), 
    .load_o     (load_buf), 
    .store_o    (store_buf),
    .rtype_o    (rtype_buf), 
    .va_o       (va_buf),
    .reg_hi_o   (reg_hi_o),
    .reg_lo_o   (reg_lo_o),
    .rd_sc_o    (rd_sc_buf),
    .reg_hwo_o  (reg_hwo_o),
    .reg_lwo_o  (reg_lwo_o),
    .sys_reg_o  (sys_reg_o),
    .set_load_o (set_load_o),
    .set_store_o (set_store_o),
    .access_ex_o (access_ex_o)
  );
  
  always @ (posedge clk_i)
    begin
      opcode_buf <= opcode_i;       // To remember current value of opcode
    end
  
  always @ (negedge clk_i)
    begin
      case ({stall_i,slip_i})       // Synthesis full case
        2'b00:
          begin
            opcode_o <= opcode_buf;     // Opcode to be used by M stage
            if (({branch_o,rtype_buf} === 2'b11) && (opcode_buf[5:0] !== `SP_JALR))
              rda_o <= 5'b11111;            // Indirect addressing of r31
            else
              rda_o <= rda_buf;   // Otherwise regular operation
            va_o <= va_buf;             // Virtual address to be used by M stage
            if (!access_ex_o)             // Load only if there is no access violation
              begin
                load_o <= load_buf;         // Delay for the M stage
                store_o <= store_buf;       // Delay for the M stage
              end
            else
              begin
                load_o <= 1'b0;     
                store_o <= 1'b0;
              end
            rd_o <= rd_buf;
            rd_sc_o <= rd_sc_buf;       // Delay for the M stage
            set_branch_o <= branch_o;
            if (!ioverflow_o)
              rtype_o <= rtype_buf;   // Set flag only if no overflow. Otherwise result should not be written
            else                      // to the reg file
              rtype_o <= 1'b0;        // If this flag is cleared the result of operation will not be written
              
          end
        2'b01:
          begin
            opcode_o <= 32'h00000000;
            load_o <= 1'b0;
            store_o <= 1'b0;
            rtype_o <= 1'b0;
            set_branch_o <= 1'b0;
          end
      endcase
    end
endmodule
