//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Execution core main file
//
module exec_core (
  clk_i, rst_i, cmd_ready_i, ins_i, data_ready_i, mmu_data_i, sys_reg_i,
  sys_pco_i, ex_branch_i, sys_eslip_i, irq_i, irq_branch_i, kernel_mode_i,
  uart_txrdy_i,
  pc_o, pc_prev_o, mmu_data_rd_o, mmu_data_we_o, mmu_data_o, va_o, 
  byte_mask_o, instrio_o, sys_reg_o, va_ex_o, ioverflow_o, set_load_o, 
  set_store_o, branch_o, stalli_o, access_ex_o, uart_wr_o, uart_data_o

);
  input clk_i;                // Input clock
  input rst_i;                // Reset input signal
  input cmd_ready_i;          // From cache controller
  input [31:0] ins_i;         // Instruction input
  input data_ready_i;         // From cache controller
  input [31:0] mmu_data_i;    // Input data from the MMU
  input [31:0] sys_reg_i;     // Input system register from CP0
  input [31:0] sys_pco_i;     // Input exception program counter from sys CP0
  input ex_branch_i;          // Exception branch signal
  input sys_eslip_i;          // Input signal from sys CP0 to slip E stage
  input irq_i;                // Signal to the I stage that exception was due to interrupt
  input irq_branch_i;         // Signal to the I stage that IRQ occured during branch delay slot
  input kernel_mode_i;        // Signal from MMU that program is in kernel mode
  input uart_txrdy_i;         // Signal from UART
  output [31:0] pc_o;         // Program counter
  output [31:0] pc_prev_o;    // Previous instruction programm counter 
  output mmu_data_rd_o;       // RD data signal for MMU
  output mmu_data_we_o;       // WE data signal for MMU
  output [31:0] mmu_data_o;   // Output data to the MMU 
  output [31:0] va_o;         // Virtual address for MMU
  output [3:0] byte_mask_o;   // Output byte mask for MMU
  output [31:0] instrio_o;    // Output opcode from the I stage to be used by CP0
  output [31:0] sys_reg_o;    // Output sys register from MTC0 instruction
  output va_ex_o;             // Exception on load/store operation occured
  output ioverflow_o;         // Arithmetic instruction overflow exception
  output set_load_o;          // To tell to the sys CP0 that exception is from LOAD instruction
  output set_store_o;         // To tell to the sys CP0 that exception is from STORE instruction
  output branch_o;            // To find out if the instruction is in branch delay slot
  output stalli_o;            // Output for MMU that I stage stalled
  output access_ex_o;         // Signal from ALU that access violation occured		
  output uart_wr_o;           // Write enable for UART
  output [7:0] uart_data_o;   // Output data for UART	 
  
// Internal wires and regs
  wire [31:0] pc_prev_w;
  wire [31:0] instr_w;
  wire [31:0] rdw_w;
  wire [4:0] rdaw_w;
  wire wen_w;
  wire [31:0] rdms_w;
  wire [31:0] rdas_w;
  wire [31:0] rde_w;
  wire [4:0] rdae_w;
  wire [31:0] pce_w;
  wire [31:0] opcodee_w;
  wire [31:0] rdm_w;
  wire [4:0] rdam_w;
  wire [31:0] rda_w;
  wire [4:0] rdaa_w;
  wire branch_w;
  wire loade_w;
  wire storee_w;
  wire [31:0] vae_w;
  wire [31:0] rs_reg_w;
  wire [31:0] rt_reg_w;
  wire [31:0] rs_bps_w;
  wire [31:0] rt_bps_w;
  wire rtypee_w;
  wire rtypem_w;
  wire stallio_w;         // Output ref to the hazard unit
  wire stallmi_w;
  wire stalleo_w;
  wire stallmo_w;
  wire slipeo_w;
  wire [31:0] rdlm_w;
  wire [4:0] rdlma_w;
  wire [5:0] opcodem_w;
  wire loadm_w;
  wire [1:0] offset01_w;
  wire writea_w;
  wire [31:0] rdl_aligned_w;
  wire stallmdu_w;
  wire [31:0] reg_hii_w;    // Input,Output reference to the A-stage
  wire [31:0] reg_loi_w;
  wire [31:0] reg_hio_w;
  wire [31:0] reg_loo_w;
  wire mul_w;
  wire [4:0] rdamul_w;
  wire stall_load_w;
  wire [31:0] reg_hieo_w;
  wire [31:0] reg_loeo_w;
  wire reg_hwe_w;
  wire reg_lwe_w;
  wire reg_hwm_w;
  wire reg_lwm_w;
  wire [31:0] rd_sc_w;
  
  assign mmu_data_rd_o = loade_w;
  assign mmu_data_we_o = storee_w;
  assign instrio_o = instr_w;
  assign pc_prev_o = pc_prev_w;
  assign stalli_o = stallio_w;
  assign branch_o = branch_w;
  
  
// Instantiate modules
i_stage istage_e (
  .clk_i          (clk_i),
  .rst_i          (rst_i), 
  .instr_i        (ins_i), 
  .cmd_ready_i    (cmd_ready_i),
  .branch_i       (branch_w),
  .pc_branch_i    (pce_w),
  .stall_i        (stallio_w),
  .sys_pco_i      (sys_pco_i),
  .ex_branch_i    (ex_branch_i),
  .irq_i          (irq_i),
  .irq_branch_i   (irq_branch_i),
  .uart_txrdy_i   (uart_txrdy_i),
  .pc_o           (pc_o), 
  .pc_prev_o      (pc_prev_w), 
  .instr_o        (instr_w),
  .uart_wr_o      (),
  .uart_data_o    ()
);

e_stage estage_e (
  .clk_i          (clk_i),
  .rst_i          (rst_i), 
  .opcode_i       (instr_w), 
  .rdw_i          (rda_w), 
  .rdaw_i         (rdaa_w), 
  .wen_i          (writea_w), 
  .rdms_i         (rdms_w), 
  .rdas_i         (rdas_w), 
  .pc_i           (pc_prev_w),
  .pc_next_i      (pc_o),
  .rs_bps_i       (rs_bps_w),
  .rt_bps_i       (rt_bps_w),
  .reg_hi_i       (reg_hio_w),
  .reg_lo_i       (reg_loo_w),
  .stall_i        (stalleo_w),
  .slip_i         (slipeo_w),
  .sys_reg_i      (sys_reg_i),
  .kernel_mode_i  (kernel_mode_i), 
  .uart_txrdy_i   (uart_txrdy_i),
  .rd_o           (rde_w), 
  .rda_o          (rdae_w),
  .pc_o           (pce_w), 
  .ioverflow_o    (ioverflow_o),
  .opcode_o       (opcodee_w),
  .branch_o       (branch_w),
  .va_ex_o        (va_ex_o), 
  .load_o         (loade_w),
  .store_o        (storee_w),
  .rtype_o        (rtypee_w), 
  .va_o           (vae_w),
  .rs_reg_o       (rs_reg_w),
  .rt_reg_o       (rt_reg_w),
  .reg_hi_o       (reg_hieo_w),
  .reg_lo_o       (reg_loeo_w),
  .rd_sc_o        (rd_sc_w),
  .reg_hwo_o      (reg_hwe_w),
  .reg_lwo_o      (reg_lwe_w),
  .sys_reg_o      (sys_reg_o),
  .set_load_o     (set_load_o),
  .set_store_o    (set_store_o),
  .set_branch_o   (),
  .access_ex_o    (access_ex_o),
  .uart_wr_o      (uart_wr_o),
  .uart_data_o    (uart_data_o)
);

m_stage mstage_e (
  .clk_i          (clk_i), 
  .rst_i          (rst_i), 
  .rd_i           (rde_w), 
  .rda_i          (rdae_w),
  .op_i           (opcodee_w[31:26]), 
  .load_i         (loade_w), 
  .store_i        (storee_w),
  .rtype_i        (rtypee_w),
  .mmu_data_i     (mmu_data_i),
  .data_rdy_i     (data_ready_i),
  .va_i           (vae_w),
  .ls_reg_i       (opcodee_w[20:16]),
  .stall_i        (stallmo_w),
  .rd_sc_i        (rd_sc_w), 
  .uart_txrdy_i   (uart_txrdy_i),
  .rd_o           (rdm_w), 
  .rda_o          (rdam_w),
  .mmu_data_o     (mmu_data_o),
  .va_o           (va_o),
  .rtype_o        (rtypem_w),
  .rdl_o          (rdlm_w),
  .rdla_o         (rdlma_w),
  .stall_o        (stallmi_w),  
  .op_o           (opcodem_w),
  .load_o         (loadm_w),
  .offset01_o     (offset01_w),
  .byte_mask_o    (byte_mask_o),
  .uart_wr_o      (),
  .uart_data_o    ()
);

a_stage astage_e (
  .clk_i          (clk_i), 
  .rst_i          (rst_i), 
  .rd_i           (rdm_w), 
  .rda_i          (rdam_w),
  .rtype_i        (rtypem_w),
  .load_i         (loadm_w),
  .opcode_i       (opcodem_w),
  .rdl_i          (rdlm_w),
  .rdla_i         (rdlma_w),
  .offset01_i     (offset01_w),
  .reg_hi_i       (reg_hii_w),
  .reg_lo_i       (reg_loi_w),
  .reg_hia_i      (reg_hieo_w),
  .reg_loa_i      (reg_loeo_w),
  .reg_hwa_i      (reg_hwe_w),
  .reg_lwa_i      (reg_lwe_w),
  .reg_hwm_i      (reg_hwm_w),
  .reg_lwm_i      (reg_lwm_w),
  .mul_i          (mul_w),
  .rdamul_i       (rdamul_w),
  .rd_o           (rda_w), 
  .rda_o          (rdaa_w),
  .write_o        (writea_w),
  .rdl_aligned_o  (rdl_aligned_w),
  .reg_hi_o       (reg_hio_w),
  .reg_lo_o       (reg_loo_w)
);

w_stage wstage_e (
  .clk_i          (clk_i),
  .rst_i          (rst_i), 
  .rd_i           (rda_w), 
  .rda_i          (rdaa_w),
  .wen_i          (writea_w),  
  .rd_o           (rdw_w), 
  .rda_o          (rdaw_w),
  .wen_o          (wen_w)
);

bypass bypass_e (
  .clk_i          (clk_i), 
  .rst_i          (rst_i), 
  .rsae_i         (instr_w[25:21]), 
  .rtae_i         (instr_w[20:16]), 
  .rdam_i         (rdae_w), 
  .rdaa_i         (rdam_w), 
  .rse_i          (rs_reg_w), 
  .rte_i          (rt_reg_w),
  .rdm_i          (rde_w), 
  .rda_i          (rdm_w),
  .rdl_aligned_i  (rdl_aligned_w),
  .rdlaa_i        (rdlma_w),
  .rtypem_i       (rtypee_w),
  .rtypea_i       (rtypem_w),
  .loada_i        (loadm_w),
  .rs_o           (rs_bps_w), 
  .rt_o           (rt_bps_w)
);

hazard hazard_e (
  .clk_i          (clk_i), 
  .rst_i          (rst_i), 
  .stallm_i       (stallmi_w),
  .load_i         (loade_w),
  .rse_i          (instr_w[25:21]),
  .rte_i          (instr_w[20:16]),
  .rlm_i          (opcodee_w[20:16]),
  .stallmdu_i     (stallmdu_w),
  .sys_eslip_i    (sys_eslip_i),
  .stalli_o       (stallio_w), 
  .stalle_o       (stalleo_w), 
  .stallm_o       (stallmo_w), 
  .stalla_o       (), 
  .stallw_o       (),
  .slipe_o        (slipeo_w),
  .stall_load_o   (stall_load_w)
);

mdu mdu_e (
  .clk_i          (clk_i),
  .rst_i          (rst_i),
  .opcode_i       (instr_w),
  .rs_i           (rs_bps_w),
  .rt_i           (rt_bps_w),
  .reg_hi_i       (reg_hio_w),
  .reg_lo_i       (reg_loo_w),
  .stall_load_i   (stall_load_w),
  .stall_o        (stallmdu_w),
  .reg_hi_o       (reg_hii_w),
  .reg_lo_o       (reg_loi_w),
  .mul_o          (mul_w),
  .rda_o          (rdamul_w),
  .reg_hiw_o      (reg_hwm_w),
  .reg_low_o      (reg_lwm_w)
);

endmodule 
