  //==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// I Stage of the MIPS32-4kp execution core
// Stage starts on positive edge of clock

`include "parms.v"
`include "opcodes_def.v"

module i_stage (
  clk_i, rst_i, instr_i, cmd_ready_i, branch_i, pc_branch_i,
  stall_i, sys_pco_i, ex_branch_i, irq_i, irq_branch_i, uart_txrdy_i,
  pc_o, pc_prev_o, instr_o, uart_wr_o, uart_data_o
  );
  
  
  input clk_i;                    // Clock input to the module
  input rst_i;                    // Reset signal - low active
  input [31:0] instr_i;           // Instruction fetched from MMU
  input cmd_ready_i;              // Input signal from MMU
  input branch_i;                 // Input specifying if branch required
  input [31:0] pc_branch_i;       // PC value from branch instruction
  input stall_i;                  // Signals that no data processing to be done
  input [31:0] sys_pco_i;         // Input exception program counter from sys CP0
  input ex_branch_i;              // Signal from sys CP0 that exception branch to be taken
  input irq_i;                    // Signal from CP0 that exception was due to interrupt
  input irq_branch_i;             // Signal from CP0 that IRQ has occured during branch delay
  input uart_txrdy_i;             // Signal from UART that it ready to Tx data
  output [31:0] pc_o;             // Program Counter - address of the next command to be fetched
  output [31:0] pc_prev_o;        // Previous value of PC counter
  output [31:0] instr_o;          // Output instruction for E stage
  output uart_wr_o;               // Write enable for UART
  output [7:0] uart_data_o;       // Output data to UART
  
  reg [31:0] pc_o;
  reg [31:0] instr_o;
  reg [31:0] pc_prev;             // This register is to create 0.5 cycle delay for the estage
  reg [31:0] pc_prev_o;
  reg branchi_delay;              // To retain branch flan when cmd_ready_i = 1'b0
  reg branchi_delay_pos;          // To shift transition of branchi_delay to the posedge 
  
  // Test registers
  reg [3:0] sw_r = 4'h0;
  reg [3:0] strap_r = 4'h0;		  
  reg [9:0] count_r = 10'h41;
 
  
  // Instantiate modules
  test_m uart_link (
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .sdata1_i             (pc_o),
    .sdata2_i             (instr_o),
    .sw_i                 (sw_r),
    .strap_i              (strap_r),
    .count_i              (count_r),
    .uart_txrdy_i         (uart_txrdy_i),
    .start_i              (rst_i),
    .uart_wr_o            (uart_wr_o),
    .uart_data_o          (uart_data_o)
  );

  
  always @ (posedge clk_i)
    begin  
      pc_prev <= pc_o;      // Remember current value of PC
      branchi_delay_pos <= branchi_delay; 
    end

  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          pc_o <= `PC_RESET;
          pc_prev_o <= `PC_RESET;
          instr_o <= `SP_SLL;         // NOP operation
          branchi_delay <= 1'b0;  
        end   
      else
        begin
          casex ({(stall_i|ex_branch_i),cmd_ready_i, branch_i})    // Synthesis full case
            3'b011:
              begin
                pc_o <= pc_branch_i;      // To take PC from branch instruction
                instr_o <= instr_i;
                pc_prev_o <= pc_prev;
              end
            3'b010:
              begin
                if (branchi_delay_pos)
                  begin
                    pc_o <= pc_branch_i;
                    branchi_delay <= 1'b0;
                  end
                else
                  pc_o <= pc_o + 4'h4;
                instr_o <= instr_i;
                pc_prev_o <= pc_prev;
              end
            3'b00x:
              begin
                instr_o <= 32'h0;         // Slip condition - insert NOP
                if (branch_i)
                  branchi_delay <= 1'b1;
              end
            3'b1xx:
              begin
                if (ex_branch_i)
                  begin
                    pc_o <= sys_pco_i;    // pc_prev_o is not updated at this stage to keep address of exception instruction for CP0
                    instr_o <= 32'h0;     // No branch delay slot for exception branch. Otherwise processor locks up in stall condition
                  end                       
              end                           
          endcase 
        end					 
    
    end
  
endmodule
