//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// MDU unit

`include "opcodes_def.v"

module mdu (
  clk_i, rst_i, opcode_i, rs_i, rt_i, reg_hi_i, reg_lo_i, stall_load_i,
  stall_o, reg_hi_o, reg_lo_o, mul_o, rda_o, reg_hiw_o, reg_low_o
  
  );
  input clk_i;                  // Input clock
  input rst_i;                  // Input reset signal
  input [31:0] opcode_i;        // Input opcode from the I stage
  input [31:0] rs_i;            // rs register from the bypass unit
  input [31:0] rt_i;            // rt register from the bypass unit
  input [31:0] reg_hi_i;        // hi register
  input [31:0] reg_lo_i;        // lo register
  input stall_load_i;           // Input stall signal from the LOAD command
  output stall_o;               // Output signal to stall main pipeline
  output [31:0] reg_hi_o;       // MSB part of the result (hi register)
  output [31:0] reg_lo_o;       // LSB part of the result (lo register)
  output mul_o;                 // Flag that current instruction is MUL
  output [4:0] rda_o;           // Address of the rd register from MUL operation
  output reg_hiw_o;             // Signal to the A stage to update hi register
  output reg_low_o;             // Signal to the A stage to update lo register
  
  reg stall_o;
  reg mul_o;
  reg [4:0] rda_o;
  reg reg_hiw_o;
  reg reg_low_o;
  
  // Internal wires
  reg [63:0] multiplicand;
  reg [63:0] product;
  reg [31:0] multiplier;
  reg [63:0] remainder;
  reg [63:0] divisor;
  reg [31:0] quotient;
  reg [5:0] counter;
  reg [31:0] reg_hi_o;
  reg [31:0] reg_lo_o;
  reg [63:0] remainder_temp;
  reg [31:0] rs_signed;
  reg [31:0] rt_signed;
  reg rs_sign;              // If zero sign is +
  reg rt_sign;              // If zero sign is +
  
  wire m_madd = (opcode_i[31:26] == `OP_SPECIAL2) && (opcode_i[5:0] == `SP2_MADD);
  wire m_maddu = (opcode_i[31:26] == `OP_SPECIAL2) && (opcode_i[5:0] == `SP2_MADDU);
  wire m_msub = (opcode_i[31:26] == `OP_SPECIAL2) && (opcode_i[5:0] == `SP2_MSUB);
  wire m_msubu = (opcode_i[31:26] == `OP_SPECIAL2) && (opcode_i[5:0] == `SP2_MSUBU);
  
  wire m_mul = (opcode_i[31:26] == `OP_SPECIAL2) && (opcode_i[5:0] == `SP2_MUL);
  wire m_mult = (opcode_i[31:26] == `OP_SPECIAL) && (opcode_i[5:0] == `SP_MULT);
  wire m_multu = (opcode_i[31:26] == `OP_SPECIAL) && (opcode_i[5:0] == `SP_MULTU);
  
  wire m_div = (opcode_i[31:26] == `OP_SPECIAL) && (opcode_i[5:0] == `SP_DIV);
  wire m_divu = (opcode_i[31:26] == `OP_SPECIAL) && (opcode_i[5:0] == `SP_DIVU);
  
  // Wire to combine operations with the same latency
  wire m_sum1 = m_madd || m_maddu || m_msub || m_msubu;
  wire m_sum2 = m_mul || m_mult || m_multu;
  wire signed_mul = m_madd || m_msub || m_mul || m_mult;
  
  always @ (posedge clk_i)                    
    begin
      if (!rst_i)
        begin
          reg_hi_o <= 32'h0;
          reg_lo_o <= 32'h0;
          stall_o <= 1'b0;
          mul_o <= 1'b0;  
        end
      else
        begin
          if (m_sum1 && (counter !== 6'd33))        // Counter logic - to specify 
            stall_o <= 1'b1;                        // duration of various mul/div cycles
          else if (m_sum2 && (counter !== 6'd32))
            stall_o <= 1'b1;
          else if ((m_div|m_divu) && (counter !== 6'd34))
            stall_o <= 1'b1;
          else
            stall_o <= 1'b0;  
      
      
          case ({(m_sum1|m_sum2),counter})             // Synthesis full case
            7'b1000000:
              begin
                if (rs_i[31] && signed_mul)    // Check sign only for signed versions of sum
                  begin         // rs is negative
                    multiplicand <= {32'h0,(32'h0 - rs_i)};
                    rs_sign <= 1'b1;
                  end
                else
                  begin       // rs is positive
                    multiplicand <= {32'h0,rs_i};
                    rs_sign <= 1'b0;
                  end
                if (rt_i[31] && signed_mul)   // Check sign only for signed versions of sum
                  begin       // rt is negative
                    multiplier <= 32'h0 - rt_i;
                    rt_sign <= 1'b1;
                  end
                else
                  begin     // rt is positive
                    multiplier <= rt_i;
                    rt_sign <= 1'b0;
                  end
                product <= 64'h0;
              end
            7'b1100001:                   // d'33
              begin
                if (m_madd|m_maddu)
                  begin
                    if ((rt_sign^rs_sign) && signed_mul)
                      {reg_hi_o,reg_lo_o} <= (64'h0 - product) + {reg_hi_i,reg_lo_i};
                    else
                     {reg_hi_o,reg_lo_o} <= product + {reg_hi_i,reg_lo_i};
                  end
                else
                  begin
                    if ((rt_sign^rs_sign) && signed_mul)
                      {reg_hi_o,reg_lo_o} <= {reg_hi_i,reg_lo_i} - (64'h0 - product);
                    else
                      {reg_hi_o,reg_lo_o} <= {reg_hi_i,reg_lo_i} - product;
                  end
              end
            default:
              begin
                if (m_sum2 && (counter === 6'd32))
                  begin
                    if (multiplier[0])
                      begin
                        if ((rt_sign^rs_sign) && signed_mul)    // Check if result should be negative                          
                          {reg_hi_o,reg_lo_o} <= 64'h0 - (product + multiplicand);
                        else
                          {reg_hi_o,reg_lo_o} <= product + multiplicand;
                      end
                    else
                      begin
                        if ((rt_sign^rs_sign) && signed_mul)
                          {reg_hi_o,reg_lo_o} <= 64'h0 - product;
                        else
                          {reg_hi_o,reg_lo_o} <= product;
                      end
                  end
                else
                  begin
                    if (multiplier[0])
                      product <= product + multiplicand;
                    multiplicand <= multiplicand << 1;
                    multiplier <= multiplier >> 1;
                  end
              end        
          endcase 
          
          if (m_mul && (counter === 6'd32))       // Logic to signal to A stage that result to be
            begin                                 // recorded into the regfile
              mul_o <= 1'b1;
              rda_o <= opcode_i[15:11];
            end
          else
            mul_o <= 1'b0;
        end
        
      if (rst_i)              // Divide logic
        begin
          case ({(m_div|m_divu),counter})        // Synthesis full case
            7'b1000000:
              begin
                if (rs_i[31] && m_div)          // Check sign only for signed version of div       
                  begin
                    rs_sign <= 1'b1;        // rs is negative number
                    rs_signed <= 32'h0 - rs_i;    // Convert negative to positive
                  end
                else
                  begin
                    rs_sign <= 1'b0;        // Otherwise rs is positive
                    rs_signed <= rs_i;
                  end
                if (rt_i[31] && m_div)      // Check sign only for signed version of div
                  begin
                    rt_sign <= 1'b1;        // rt is negative number
                    rt_signed <= 32'h0 - rt_i;      // Convert negative to positive
                  end
                else
                  begin
                    rt_sign <= 1'b0;        // Otherwise rt is positive
                    rt_signed <= rt_i;
                  end                  
              end
            7'b1000001:                           // Initialize registers
              begin
                remainder <= {32'h0,rs_signed};
                divisor <= {1'b0,rt_signed,31'h0};    // Divisor needs to be shifted right one bit before first operation
                quotient <= 32'h0;
              end
            7'b1100010:                       //6'd34
              begin
                if ((rt_sign^rs_sign) && (!m_div))   // Check sign only for signed version of div
                  reg_lo_o <= 32'h0 - quotient;   // Resulting number is negative
                else
                  reg_lo_o <= quotient;           // Result is positive
                  
                if (rs_sign && m_div)             // Remainder sign = dividend sign. Check sign only for signed version of div
                  reg_hi_o <= 32'h0 - remainder[31:0];
                else
                  reg_hi_o <= remainder[31:0];
              end
            default:
              begin
                if (!remainder_temp[63])    // If zero result is positive
                  begin
                    quotient[31:0] <= {quotient[30:0],1'b1};
                    remainder <= remainder_temp;
                    divisor <= divisor >> 1;
                  end
                else
                  begin
                    quotient[31:0] <= {quotient[30:0],1'b0}; 
                    divisor <= divisor >> 1;              
                  end
              end
          endcase
          
        end 
        
        if (({m_sum1,counter} === 7'b1100001) || ({m_sum2,counter} === 7'b1100000) || ({(m_div|m_divu),counter} === 7'b1100010))
          {reg_hiw_o,reg_low_o} <= 2'b11;
        else
          {reg_hiw_o,reg_low_o} <= 2'b00;

    end
    
  always @ (negedge clk_i)
    begin
      if ((!stall_o)||(stall_load_i))
        counter <= 6'b000000;
      else 
        counter <= counter + 1'b1;
      
      remainder_temp <= remainder - divisor;
        
    end
  
endmodule