//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Memory controller top module

module mem_ctrl (
  clk_i, rst_i, eb_avalid_i, eb_instr_i, eb_write_i,
  eb_burst_i, eb_bfirst_i, eb_blast_i, eb_blen_i,
  eb_be_i, eb_adr_i, eb_data_i, uart_txrdy_i, uart_rxrdy_i, uart_data_i,
  init_done_i, sw_i, btn_i,
  eb_ardy_o, eb_data_o, eb_rdval_o, eb_wdrdy_o, eb_rberr_o,
  eb_wberr_o, uart_rd_o, uart_wr_o, uart_data_o, led_o
  );
  
  input clk_i;                // Input clock
  input rst_i;                // Input reset signal
  input eb_avalid_i;          // Address bus and access type lines are valid
  input eb_instr_i;           // Transaction is an instruction fetch
  input eb_write_i;           // Current transaction is write
  input eb_burst_i;           // Current transaction is part of a cache fill or a write burst
  input eb_bfirst_i;          // Beginning of the burst
  input eb_blast_i;           // End of the burst
  input [1:0] eb_blen_i;      // Burst length
  input [3:0] eb_be_i;        // Byte enable signal
  input [31:0] eb_adr_i;      // Address bus
  input [31:0] eb_data_i;     // Data bus	   
  input uart_txrdy_i;         // Input Tx ready signal from UART
  input uart_rxrdy_i;         // Input Rx ready signal from UART
  input [7:0] uart_data_i;    // Input data from UART
  input init_done_i;          // One when initialization of memory by UART is done
  input [3:0] sw_i;           // Input GPIO switch
  input [3:0] btn_i;          // Input buttons
  output eb_ardy_o;           // Indicates whether the target is ready for a new address
  output [31:0] eb_data_o;    // Output data bus
  output eb_rdval_o;          // Target is driving read data
  output eb_wdrdy_o;          // Target of  write is ready
  output eb_rberr_o;          // Bus error for read transaction
  output eb_wberr_o;          // Bus error for write transaction
  output uart_rd_o;           // Read UART signal
  output uart_wr_o;           // Write UART signal
  output [7:0] uart_data_o;   // Output data to UART  
  output [3:0] led_o;         // Output data to drive LEDs

  reg eb_ardy_o;
  reg eb_rdval_o;
  reg [31:0] eb_data_o;
  reg eb_rberr_o;
  reg eb_wdrdy_o;
  reg eb_wberr_o;
  reg uart_rd_o;
  reg uart_wr_o;										  
  reg [3:0] led_o;							
  reg [7:0] uart_data_o;
  reg [7:0] uart_data_neg;
  reg uart_rd_neg;
  reg uart_wr_neg;
  reg [3:0] counter;  							  
  reg [1:0] counter1;
  reg [1:0] byte_counter;		   
  reg sram_we;							
  reg [31:0] counter_init; 
  reg [7:0] counter_tx;

  reg [3:0] opcode;           // Operation to be performed by memory controller
  reg [3:0] opcode_buf;       // To shift transition of opcode to the posedge
  reg [31:0] eb_adri_pos;
  reg [31:0] eb_adri_neg;     // Two address buffers to delay address for one clock
  reg eb_blast_pos;           // To shift 0.5 clk
  reg eb_blast_neg;           // To shift 0.5 clk further			   
  reg [12:0] data_ptr;        // Data pointer for UART to init MIPS's memory  
  reg [12:0] data_ptr_dly;    // One clock delayed data pointer
  reg uart_rxrdy_neg;         // Delay for 0.5 clk
  reg uart_rxrdy_pos;         // Delay for 0.5 clk further
  reg [7:0] tmp_reg;          // General purpose register
  reg [7:0] uart_datai_neg;    // To shift transition of input UART data to the negedge
  
  wire [31:0] eb_datai_w;
  wire [31:0] eb_datao_w;
  wire [3:0] eb_be_w;			  			
  wire [12:0] eb_adr_w;
  reg [3:0] eb_be_r;
  					
  assign eb_be_w = (init_done_i) ? eb_be_i : eb_be_r;		
  assign eb_datai_w = (init_done_i) ? eb_data_i : {uart_datai_neg,uart_datai_neg,uart_datai_neg,uart_datai_neg};
  assign eb_adr_w = (init_done_i) ? {eb_adr_i[31:29],2'b00,eb_adr_i[9:2]} : data_ptr_dly; 
// Instantiate modules
  memory sram_m (
    .clk_i                (clk_i), 
    .we_i                 (sram_we),  
    .data_i               (eb_datai_w), 
    .be_i                 (eb_be_w), 
    .adr_i                (eb_adr_w),
    .data_o               (eb_datao_w)
  ); 
  
  /*
  // Test logic
  reg [3:0] sw_r = 4'h0;
  reg [3:0] strap_r = 4'h0;
  reg [9:0] count_r = 10'h41; 
  
    test_m uart_link (
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .sdata1_i             (eb_data_i),
    .sdata2_i             (eb_adr_i),
    .sw_i                 (sw_r),
    .strap_i              (strap_r),
    .count_i              (count_r),
    .uart_txrdy_i         (uart_txrdy_i),
    .start_i              (init_done_i),
    .uart_wr_o            (uart_wr_o),
    .uart_data_o          (uart_data_o)
  );                                 
      
  */    
  always @ (posedge clk_i)
    begin
      opcode_buf <= opcode;
      eb_adri_pos <= eb_adri_neg;
      eb_blast_pos <= eb_blast_neg;
      uart_rd_o <= uart_rd_neg;
      uart_wr_o <= uart_wr_neg;
      uart_data_o <= uart_data_neg;						  
      uart_rxrdy_pos <= uart_rxrdy_neg;
    end 
 

  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          eb_ardy_o <= 1'b0;
          eb_rdval_o <= 1'b0;
          eb_wdrdy_o <= 1'b0;
          eb_rberr_o <= 1'b0;
          eb_wberr_o <= 1'b0;
          opcode <= 4'h0;	   
          uart_rd_neg <= 1'b0;
          counter <= 4'h0;	
          byte_counter <= 2'b00;
          led_o <= 4'h0;													
          data_ptr <= 13'b0;   
          sram_we <= 1'b0;	
          counter_init <= 32'h0; 
          counter_tx <= 8'h0;													 
          counter1 <= 2'b00;
        end
      else if (init_done_i && rst_i)                 // When initialization is done
        begin
          casex ({eb_avalid_i, eb_write_i, eb_burst_i, eb_instr_i})      // Synthesis full case
            4'b1001: opcode <= 4'h1;           // Single instruction fetch
            4'b1011: opcode <= 4'h2;           // Burst read for instruction - 4 words
            4'b1000: opcode <= 4'h3;           // Single data word read
            4'b110x: opcode <= 4'h4;           // Single data word write
          endcase
          
          case (opcode_buf)             // Synthesis full case
            4'h0:                       // NOP
              begin
                eb_rdval_o <= 1'b0;
                eb_ardy_o <= 1'b1;
                eb_wdrdy_o <= 1'b0;
                counter <= 4'h0;
                uart_rd_neg <= 1'b0;
                uart_wr_neg <= 1'b0;   
                sram_we <= 1'b0;
              end
            4'h1, 4'h3:                       // Single word read
              begin 
                if (eb_adri_pos[11:10] === 2'b00)         // Memory access
                  begin									
                    eb_data_o <= eb_datao_w;        // Read from sram block memory
                    eb_rdval_o <= 1'b1;
                    eb_ardy_o <= 1'b0;
                    opcode <= 4'h0;
                  end
                else if (eb_adri_pos[11:10] === 2'b01)                     // UART access
                  begin
                    counter <= counter + 1'b1;
                    if (counter === 4'h0)
                      uart_rd_neg <= 1'b1;
                    else 
                      uart_rd_neg <= 1'b0;								   
                    if (counter === 4'h2)        // At this time data is available from UART
                      begin
                        eb_data_o[7:0] <= uart_data_i[7:0];
                        eb_rdval_o <= 1'b1;
                        eb_ardy_o <= 1'b0;
                        opcode <= 4'h0;
                      end 
                  end
                else if (eb_adri_pos[11:10] == 2'b10)                // Read status of txrdy and rxrdy lines
                  begin
                    eb_data_o[1:0] <= {uart_txrdy_i,uart_rxrdy_i};
                    eb_rdval_o <= 1'b1;
                    eb_ardy_o <= 1'b0;
                    opcode <= 4'h0;
                  end
              end
            4'h2:                           // Burst read
              begin
    /*            eb_data_o <= {data[{eb_adri_pos[31:29],eb_adri_pos[8:2],2'b11}],
                              data[{eb_adri_pos[31:29],eb_adri_pos[8:2],2'b10}],
                              data[{eb_adri_pos[31:29],eb_adri_pos[8:2],2'b01}],
                              data[{eb_adri_pos[31:29],eb_adri_pos[8:2],2'b00}]};
                eb_rdval_o <= 1'b1;
                eb_ardy_o <= 1'b1;
                if (eb_blast_pos)
                  begin
                    opcode <= 4'h0;
                    eb_ardy_o <= 1'b0;
                  end             */   
              end
            4'h4:                       // Single word write
              begin		
                if (eb_adri_pos[11:10] === 2'b00)                // Memory write
                  begin
                    sram_we <= 1'b1;
                    eb_wdrdy_o <= 1'b1;
                    opcode <= 4'h0;
                  end
                else if (eb_adri_pos[11:10] === 2'b10)            // LED write
                  begin
                    led_o[3:0] <= eb_data_i[3:0];
                    eb_wdrdy_o <= 1'b1;
                    opcode <= 4'h0;
                  end			   
                else if (eb_adri_pos[11:10] === 2'b01)            // UART write
                  begin
                    counter <= counter + 1'b1;
                    if (counter === 4'h0)
                      begin
                        uart_wr_neg <= 1'b1;
                        uart_data_neg <= eb_data_i[7:0];        // Write data
                      end
                    else 
                      begin
                        uart_wr_neg <= 1'b0;
                        eb_wdrdy_o <= 1'b1;
                        opcode <= 4'h0;
                      end
                  end
              end
          endcase
        end	        															  
      else if (!init_done_i && rst_i)                     // In this mode UART initializes MIPS's memory
        begin                  // MIPS starts at 0x1FC00000 after reset
          counter1 <= counter1 + 1'b1;						
          if (uart_rxrdy_i)
            begin
              if (counter1 === 2'b00)
                uart_rd_neg <= 1'b1;
              else
                uart_rd_neg <= 1'b0; 
            end							 
          if ((uart_rxrdy_pos === 1'b1) && (counter1 === 2'b10))
            begin				
              byte_counter <= byte_counter + 1'b1;
              sram_we <= 1'b1;
              if (byte_counter == 2'b00)			
                eb_be_r <= 4'b0001;
              else if (byte_counter == 2'b01)
                eb_be_r <= 4'b0010;
              else if (byte_counter == 2'b10)
                eb_be_r <= 4'b0100;
              else if (byte_counter == 2'b11)		
                begin
                  eb_be_r <= 4'b1000;
                  data_ptr <= data_ptr + 1'b1;
                end
            end  			  
          else 
            begin
              sram_we <= 1'b0;
            end
          led_o <= eb_datao_w[3:0];
          
        end  

      eb_blast_neg <= eb_blast_i;
      eb_adri_neg <= eb_adr_i;			
      uart_rxrdy_neg <= uart_rxrdy_i;											
      uart_datai_neg <= uart_data_i;							
      data_ptr_dly <= data_ptr;
    end  
endmodule