//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top unit for the MIPS processor architecture 
// implemented MIPS32-4Kp core
// This core is always Little Endian
//

module mips_top (
  clk_i, cold_rst_i, nmi_i, warm_rst_i, int_i, mergemode_i,
  simplebe_i, eb_ardy_i, eb_data_i, eb_rdval_i, eb_wdrdy_i,
  eb_rberr_i, eb_wberr_i, eb_ewbe_i, eb_sblock_i, uart_txrdy_i,
  clk_o, erl_o, exl_o, rp_o, sleep_o, timerint_o, eb_avalid_o,
  eb_instr_o, eb_write_o, eb_burst_o, eb_bfirst_o, eb_blast_o,
  eb_blen_o, eb_be_o, eb_adr_o, eb_data_o, eb_wwbe_o, uart_wr_o,
  uart_data_o
   );
   	    	  
   input clk_i;               // External clock source
   input cold_rst_i;          // External cold reset signal
   input nmi_i;               // Non maskable interrupt input
   input warm_rst_i;          // Warm reset input signal
   input [5:0] int_i;         // Active high interrupt pins
   input [1:0] mergemode_i;   // Merge mode 
   input [1:0] simplebe_i;    // Mode of external bus transaction
   input eb_ardy_i;           // Indication whether the target is ready for a new address
   input [31:0] eb_data_i;    // Data input from external memory
   input eb_rdval_i;          // Indicates that target is driving read data
   input eb_wdrdy_i;          // Indicates that the target of a write is ready
   input eb_rberr_i;          // Bus error indication for read transaction
   input eb_wberr_i;          // Bus error for write transaction
   input eb_ewbe_i;           // Indicates that any external write buffers are empty
   input eb_sblock_i;         // Sub block ordering is used	   
   input uart_txrdy_i;        // Signal from UART
   output clk_o;              // Output clock signal
   output erl_o;              // State of the ERL bit in the CP0 status register
   output exl_o;              // State of the EXL bit in the CP0 status register
   output rp_o;               // State of the RP bit in the CP0 status register
   output sleep_o;            // Sleep mode
   output timerint_o;         // Timer interrupt output
   output eb_avalid_o;        // Address bus and access types lines are valid
   output eb_instr_o;         // Indicates that transaction is an instruction fetch
   output eb_write_o;         // Indicates that current transaction is write
   output eb_burst_o;         // Current transaction is part of a cache fill or a write burst
   output eb_bfirst_o;        // Beginning of the burst
   output eb_blast_o;         // End of the burst
   output [1:0] eb_blen_o;    // Burst length
   output [3:0] eb_be_o;      // Byte enable signal
   output [31:0] eb_adr_o;    // Address bus for external memory
   output [31:0] eb_data_o;   // Data output to external memory
   output eb_wwbe_o;          // Core is waiting for external write buffers to empty
   output uart_wr_o;          // Write enable signal for UART    - Test signal
   output [7:0] uart_data_o;  // Data to be sent to UART - Test signal 
   
   
   // Internal wires and regs
   wire cmd_ready_w;
   wire data_ready_w;
   wire [31:0] ins_w;
   wire [31:0] pc_w;
   wire [31:0] pc_prev_w;
   wire mmu_data_rd_w;
   wire mmu_data_we_w;
   wire [31:0] mmu_datai_w;
   wire [31:0] mmu_datao_w;
   wire [31:0] vae_w;
   wire [3:0] byte_mask_w;
   wire [31:0] instrio_w;
   wire [31:0] sys_rego_w;        // I/O relative to the sys_cp module
   wire [31:0] sys_regi_w;        // I/O relative to the sys_cp module
   wire va_ex_w;
   wire ioverflow_w;
   wire set_load_w;
   wire set_store_w;
   wire [31:0] sys_pco_w;
   wire ex_branch_w;
   wire sys_eslip_w;
   wire irq_w;
   wire irq_branch_w;
   wire [31:0] r12_status_w;
   wire [31:0] r23_debug_w;
   wire [31:0] r16_config_w;
   wire [31:0] r16_config1_w;
   wire [31:0] datami_w;            // I/O relative to BIU
   wire [31:0] datamo_w;
   wire [31:0] pham_w;
   wire rdi1_w;
   wire rdi1_ready_w;
   wire rdi4_w;
   wire rdi4_ready_w;
   wire rdd1_w;
   wire rdd1_ready_w;
   wire stalli_w;
   wire wrd1_w;
   wire wrd1_done_w;
   wire [3:0] byte_masko_w;
   wire eb_rberr_w;
   wire eb_wberr_w;
   wire kernel_mode_w;
   wire access_ex_w;

   // Modules instantiation
   exec_core exec_core_top (        // Execution core
     .clk_i           (clk_i), 
     .rst_i           (cold_rst_i), 
     .cmd_ready_i     (cmd_ready_w), 
     .ins_i           (ins_w),
     .data_ready_i    (data_ready_w),
     .mmu_data_i      (mmu_datao_w),
     .sys_reg_i       (sys_rego_w),  
     .sys_pco_i       (sys_pco_w),
     .ex_branch_i     (ex_branch_w),
     .sys_eslip_i     (sys_eslip_w),
     .irq_i           (irq_w),
     .irq_branch_i    (irq_branch_w),
     .kernel_mode_i   (kernel_mode_w),			 
     .uart_txrdy_i    (uart_txrdy_i),
     .pc_o            (pc_w),
     .pc_prev_o       (pc_prev_w),
     .mmu_data_rd_o   (mmu_data_rd_w), 
     .mmu_data_we_o   (mmu_data_we_w), 
     .mmu_data_o      (mmu_datai_w),
     .va_o            (vae_w),
     .byte_mask_o     (byte_mask_w),
     .instrio_o       (instrio_w),
     .sys_reg_o       (sys_regi_w),
     .va_ex_o         (va_ex_w),
     .ioverflow_o     (ioverflow_w),
     .set_load_o      (set_load_w),
     .set_store_o     (set_store_w),
     .branch_o        (branch_w),
     .stalli_o        (stalli_w),
     .access_ex_o     (access_ex_w),
     .uart_wr_o       (uart_wr_o),	
     .uart_data_o     (uart_data_o)
   );
   
   mmu_core mmu_core_top (          // Memory Management Unit 
     .clk_i           (clk_i), 
     .rst_i           (cold_rst_i), 
     .pc_i            (pc_w),
     .ext_data_i      (datamo_w),
     .dwe_i           (mmu_data_we_w),
     .drd_i           (mmu_data_rd_w),
     .ms_data_i       (mmu_datai_w),
     .va_i            (vae_w),
     .byte_mask_i     (byte_mask_w),
     .r12_status_i    (r12_status_w),
     .r23_debug_i     (r23_debug_w),
     .r16_config_i    (r16_config_w),
     .r16_config1_i   (r16_config1_w),
     .rdi1_ready_i    (rdi1_ready_w),
     .rdi4_ready_i    (rdi4_ready_w),
     .rdd1_ready_i    (rdd1_ready_w),
     .stalli_i        (stalli_w),
     .wrd1_done_i     (wrd1_done_w),
     .ex_branch_i     (ex_branch_w),
     .cmd_ready_o     (cmd_ready_w), 
     .ins_o           (ins_w),
     .ext_data_o      (datami_w),
     .data_ready_o    (data_ready_w),
     .ms_data_o       (mmu_datao_w),
     .pha_o           (pham_w),
     .rdi1_o          (rdi1_w),
     .rdi4_o          (rdi4_w),
     .rdd1_o          (rdd1_w),
     .wrd1_o          (wrd1_w),
     .byte_mask_o     (byte_masko_w),
     .data_access_o   (data_access_w),
     .kernel_mode_o   (kernel_mode_w)
   
   );
   biu biu_top (
     .clk_i           (clk_i),
     .cold_rst_i      (cold_rst_i),
     .dataex_i        (eb_data_i),
     .datam_i         (datami_w),
     .pham_i          (pham_w),
     .rdi1_i          (rdi1_w),
     .rdi4_i          (rdi4_w),
     .rdd1_i          (rdd1_w),
     .eb_ardy_i       (eb_ardy_i),
     .eb_rdval_i      (eb_rdval_i),
     .eb_rberr_i      (eb_rberr_i),
     .wrd1_i          (wrd1_w), 
     .byte_mask_i     (byte_masko_w),
     .eb_wdrdy_i      (eb_wdrdy_i),
     .eb_wberr_i      (eb_wberr_i),
     .uart_txrdy_i    (uart_txrdy_i),
     .dataex_o        (eb_data_o),
     .adrex_o         (eb_adr_o),
     .datam_o         (datamo_w),
     .rdi1_ready_o    (rdi1_ready_w),
     .rdi4_ready_o    (rdi4_ready_w),
     .eb_instr_o      (eb_instr_o),
     .eb_be_o         (eb_be_o),
     .eb_avalid_o     (eb_avalid_o),
     .eb_write_o      (eb_write_o),
     .eb_rberr_o      (eb_rberr_w),
     .eb_burst_o      (eb_burst_o),
     .eb_bfirst_o     (eb_bfirst_o),
     .eb_blast_o      (eb_blast_o),
     .rdd1_ready_o    (rdd1_ready_w),
     .wrd1_done_o     (wrd1_done_w),
     .eb_wberr_o      (eb_wberr_w),
     .uart_wr_o       (),
     .uart_data_o     ()
   );
  
   sys_cp sys_cp_top (
     .clk_i           (clk_i),
     .cold_rst_i      (cold_rst_i),
     .si_simplebe_i   (simplebe_i),
     .mergemode_i     (mergemode_i),
     .eb_sblock_i     (eb_sblock_i),
     .int_i           (int_i),
     .sw_int_i        (),
     .instrio_i       (instrio_w),
     .sys_reg_i       (sys_regi_w),
     .va_ex_i         (va_ex_w),
     .ioverflow_i     (ioverflow_w),
     .set_load_i      (set_load_w),
     .set_store_i     (set_store_w),
     .branch_i        (branch_w),
     .pc_i            (pc_prev_w),
     .cmd_ready_i     (cmd_ready_w),
     .pc_next_i       (pc_w),
     .eb_rberr_i      (eb_rberr_w),
     .eb_wberr_i      (eb_wberr_w),
     .data_access_i   (data_access_w),
     .access_ex_i     (access_ex_w),
     .va_i            (vae_w),
     .kernel_mode_i   (kernel_mode_w),
     .sys_reg_o       (sys_rego_w),
     .pc_o            (sys_pco_w),
     .ex_branch_o     (ex_branch_w),
     .sys_eslip_o     (sys_eslip_w),
     .irq_o           (irq_w),
     .irq_branch_o    (irq_branch_w),
     .r12_status_o    (r12_status_w),
     .r23_debug_o     (r23_debug_w),
     .r16_config_o    (r16_config_w),
     .r16_config1_o   (r16_config1_w)
   );
   
endmodule