//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Memory Management Unit top file
//
module mmu_core (
  clk_i, rst_i, pc_i, ext_data_i, dwe_i, drd_i, ms_data_i, va_i,
  byte_mask_i, r12_status_i, r23_debug_i, r16_config_i, r16_config1_i,
  rdi1_ready_i, rdi4_ready_i, rdd1_ready_i,stalli_i, wrd1_done_i,
  ex_branch_i,
  cmd_ready_o, ins_o, ext_data_o, data_ready_o, ms_data_o, pha_o, rdi1_o,
  rdi4_o, rdd1_o, wrd1_o, byte_mask_o, data_access_o, kernel_mode_o
  );
  input clk_i;                // Input clock
  input rst_i;                // Reset signal
  input [31:0] pc_i;          // Program Counter
  input [31:0] ext_data_i;    // Input from external memory
  input dwe_i;                // Write Enable from the M stage (changes on the negedge)
  input drd_i;                // Read signal from the M stage (changes on the negedge)
  input [31:0] ms_data_i;     // Input data from the M stage
  input [31:0] va_i;          // Input virtual address from the M stage
  input [3:0] byte_mask_i;    // Byte mask, if bit is set corresponding byte to be written
  input [31:0] r12_status_i;  // Input status register from CP0
                              // UM[4], ERL[2], EXL[1] 
  input [31:0] r23_debug_i;   // Input debug register from CP0
                              // DM[30]
  input [31:0] r16_config_i;  // Input config register from CP0
  input [31:0] r16_config1_i; // Input config1 register from CP0
  input rdi1_ready_i;         // Signal from BIU that one instruction is read
  input rdi4_ready_i;         // Signal from BIU that four instructions are being read
  input rdd1_ready_i;         // Signal from BIU that one data word is read
  input stalli_i;             // Input to signal that I stage of exec core is stalled
  input wrd1_done_i;          // Signal from BIU that write cycle is completed
  input ex_branch_i;          // Signal from CP0 that exception occurs
  output cmd_ready_o;         // Signal to fetch command
  output [31:0] ins_o;        // Output instruction
  output [31:0] ext_data_o;   // Output to external memory
  output data_ready_o;        // Signal to the M stage of exec_core
  output [31:0] ms_data_o;    // Output data to the M stage
  output [31:0] pha_o;        // Physical address output
  output rdi1_o;              // Request to fetch one instruction from BIU
  output rdi4_o;              // Request to fetch four instructions from BIU
  output rdd1_o;              // Request to read one data word from BIU
  output wrd1_o;              // Request to write one data word for BIU
  output [3:0] byte_mask_o;   // Output byte mask signal for BIU
  output data_access_o;       // Signal to CP0 that data access is ongoing
  output kernel_mode_o;       // When = 1 program is in kernel mode
  
  assign byte_mask_o = byte_mask_i;
  assign ext_data_o = ms_data_i;
  
  reg [31:0] ins_o;
  reg [31:0] ms_data_o;
  reg rdi4_o;
  reg rdd1_o;
  reg wrd1_o;
  reg rdi1_rq;                // Request for one instruction to be fetched
  
  reg [31:0] pha_ins;         // Physical address for instruction fetch
  reg [31:0] pha_d;           // Physical address for data access
         
  
  reg dcache_hit;             // Register to specify if data cache hit occurs
  reg icache_hit;             // Register to specify if instruction cache hit occurs
  reg data_w;                 // Register to specify start of dcache write cycle
  reg data_r;                 // Register to specify start of dcache read cycle
  reg dataw_neg;              // Register to shift transition of data_w to the negedge
  reg rw_flag;                // Flag indicating that read operation follows directly write
  reg rw_flagn;               // Copy of rw_flag syncronized on the negedge of main clock
   
  reg data_access_buf;        // To shift transition of dcache_access to the negedge
  reg [31:0] pc_i_buf;        // To shift transition of pc_i to the posedge
  reg [31:0] va_i_buf;        // To shift transition of va_i to the posedge
  reg [31:0] pcim;            // Memorized value of pc_i to be used in cache refill operation
  reg [31:0] pha_insm;        // Memorized value of pha_ins
  reg [3:0] bus_counter;      // Bus counter
  reg [3:0] bus_counter_buf;  // To shift transition of bus_counter to 0.5 clk
  reg [1:0] word_align;       // Used in burst transactions to align words
  reg [1:0] word_align_neg;   // Used with word_align to increment word count during burst transactions
  reg rdd1_readyi_neg;        // To shift transition of rdd1_ready_i to the negedge
  reg wrd1_donei_neg;         // To shift transition of wrd1_done_i to the negedge
  reg data_readyo_pos;        // To shift transition of data_ready_o to the posedge
  reg [1:0] wr_count;         // This register required when there are two consequent write commands


  assign kernel_mode_o =  (~r12_status_i[4] | r12_status_i[2] | r12_status_i[1]) & ~r23_debug_i[30];
  wire data_access = (data_w | data_r) && ~rdi1_o && ~rdi4_o;     // Start data access after current instruction
                                                                  // fetch is completed. Changes on posedge
  wire data_access_o = data_access;
  
  wire ikucache = ~pc_i_buf[31] && ~r16_config_i[27] && r16_config_i[26] && r16_config_i[25];
  wire ik0cache = pc_i_buf[31] && ~pc_i_buf[30] && ~pc_i_buf[29] && ~r16_config_i[2] && r16_config_i[1] && r16_config_i[0];
  // kseg1 is always uncacheable
  wire ik23cache = pc_i_buf[31] && pc_i_buf[30] && ~r16_config_i[30] && r16_config_i[29] && r16_config_i[28];
 // wire icacheable = ikucache || ik0cache || ik23cache;      // If one then cacheable memory is being accessed
  reg icacheable = 1'b0;        // Cache is not implemented
  wire dkucache = ~va_i_buf[31] && ~r16_config_i[27] && r16_config_i[26] && r16_config_i[25];
  wire dk0cache = va_i_buf[31] && ~va_i_buf[30] && ~va_i_buf[29] && ~r16_config_i[2] && r16_config_i[1] && r16_config_i[0];
  // kseg1 is always uncacheable
  wire dk23cache = va_i_buf[31] && va_i_buf[30] && ~r16_config_i[30] && r16_config_i[29] && r16_config_i[28];
//  wire dcacheable = dkucache || dk0cache || dk23cache;
  reg dcacheable = 1'b0;        // Dcache is not implemented

  
  assign cmd_ready_o = icache_hit || rdi1_ready_i || (data_readyo_pos & ~rw_flagn);  // Changes on the posedge
  assign data_ready_o = rdd1_readyi_neg;            // Changes on the negedge
  
  assign pha_o = (data_access) ? pha_d : pha_ins;
  assign rdi1_o = rdi1_rq && ~data_readyo_pos && ~ex_branch_i;    // ~ex_branch_i - Do not request bus access during exception cycle

   
  always @ (rdi1_ready_i or ext_data_i)        // Instruction cache fetch
    begin
      icache_hit = 1'b0;      // Cache miss occured
        
      if (rdi1_ready_i)           // Access instruction
        ins_o = ext_data_i;
    end

  always @ (rst_i or icacheable or data_access or rdi1_ready_i)                             // Uncacheable instruction fetch
    begin 
      if (!rst_i)
        rdi1_rq = 1'b0;                                 // Or cacheable instruction fetch with lock bit set in the tag
      else if (~icacheable && ~data_access)     
        begin
          if (!rdi1_ready_i)
            rdi1_rq = 1'b1;
          else
            rdi1_rq = 1'b0;
        end
      else
        rdi1_rq = 1'b0;
    end
    
  always @ (rst_i)                            // Cached instruction access with cache miss for unlocked locations
    begin
      if (!rst_i)
        rdi4_o = 1'b0;
    end               
    
    
  always @ (rdd1_ready_i or ext_data_i)      // Data cache fetch
    begin
        dcache_hit = 1'b0;
        
      if (rdd1_ready_i)     // Read one data word
        ms_data_o = ext_data_i;
    end
    
  always @ (rst_i or dcacheable or data_access or rdd1_ready_i or data_r or data_w or rdi4_o or wrd1_done_i or data_w)                    // Uncacheable data read and write
    begin
      if (!rst_i)
        begin
          rdd1_o = 1'b0;
          wrd1_o = 1'b0;
        end
      else if (~dcacheable && data_access)
        begin
          if (!rdd1_ready_i && data_r && ~data_w && ~rdi4_o)        // This is processing for the data read cycle
            rdd1_o = 1'b1;                               // When read and write occur at the same time
          else                                           // write has higher priority. rdi4 also has higher priority
            rdd1_o = 1'b0;
            
          if (!wrd1_done_i && data_w)       // This is processing for the data write cycle
            wrd1_o = 1'b1;
          else
            wrd1_o = 1'b0;
              
        end
      else
        begin
          rdd1_o = 1'b0;
          wrd1_o = 1'b0;
        end
    end  
    

  always @ (posedge clk_i)          // Address translation
    begin
      casex ({va_i[31:30],r12_status_i[2]})       // Synthesis full case
        3'b0x1: pha_d <= va_i;                    // kuseg with ERL = 1
        3'b0x0: pha_d <= {(va_i[31:28] + 4'h4),va_i[27:0]};  // kuseg with ERL = 0
        3'b10x: pha_d <= {3'b000,va_i[28:0]};                // kseg0 and kseg1
        3'b11x: pha_d <= va_i;                               // kseg2 and kseg3
      endcase
      
      casex ({pc_i[31:30],r12_status_i[2]})       // Synthesis full case
        3'b0x1: pha_ins <= pc_i;                    // kuseg with ERL = 1
        3'b0x0: pha_ins <= {(pc_i[31:28] + 4'h4),pc_i[27:0]};     // kuseg with ERL = 0
        3'b10x: pha_ins <= {3'b000,pc_i[28:0]};                   // kseg0 and kseg1
        3'b11x: pha_ins <= pc_i;                                  // kseg2 and kseg3 
      endcase
    end
      
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          data_w <= 1'b0;
          data_r <= 1'b0;
          wr_count <= 2'b00;
        end  
      else
        begin
          if (dwe_i)
            wr_count <= wr_count + 1'b1;
          else if (wrd1_donei_neg)
            wr_count <= wr_count - 1'b1;
            
          casex ({drd_i, rdd1_readyi_neg})        // Synthesis full case
            2'b10: data_r <= 1'b1;
            2'bx1: data_r <= 1'b0;
          endcase
          
          if ((wr_count > 2'b00) && ~wrd1_donei_neg)
            data_w <= 1'b1;
          else if ((wr_count != 2'b10) && wrd1_donei_neg)
            data_w <= 1'b0;
        end
        
      pc_i_buf <= pc_i;   
      va_i_buf <= va_i;  
      bus_counter_buf <= bus_counter;
      word_align_neg <= word_align + 1'b1;
      
      data_readyo_pos <= data_ready_o;
      
    end
    
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          rw_flagn <= 1'b0;
        end  
      else
        begin
          data_access_buf <= data_access;
          if (rdi4_ready_i)
            bus_counter <= bus_counter + 1'b1;
          else
            bus_counter <= 8'h00;
            
          rw_flagn <= rw_flag;
        end
      rdd1_readyi_neg <= rdd1_ready_i;
      wrd1_donei_neg <= wrd1_done_i;
      dataw_neg <= data_w;
    end
   
   always @ (posedge rdi4_ready_i)
     begin
       pcim <= pc_i;
       pha_insm <= pha_ins;
     end  
     
  always @ (posedge rdd1_o)
    begin
      if (dataw_neg)
        rw_flag <= 1'b1;      // Read after Write happened
      else
        rw_flag <= 1'b0;
    end 
    
endmodule