//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// M Stage of the MIPS32-4kp execution core
//
//
`include "opcodes_def.v"

module m_stage (
  clk_i, rst_i, rd_i, rda_i, op_i, load_i, store_i, rtype_i,
  mmu_data_i, data_rdy_i, va_i, ls_reg_i, stall_i, rd_sc_i,
  uart_txrdy_i,
  rd_o, rda_o, mmu_data_o, va_o,
  rtype_o, rdl_o, rdla_o, stall_o, op_o, load_o, offset01_o,
  byte_mask_o, uart_wr_o, uart_data_o
  );
  input clk_i;                // Input clock
  input rst_i;                // Input Reset signal
  input [31:0] rd_i;          // Input result of ALU 
  input [4:0] rda_i;          // Input address of RD
  input [5:0] op_i;           // Opcode input
  input load_i;               // Load to be done
  input store_i;              // Store to be done
  input rtype_i;              // Instruction of R-type
  input [31:0] mmu_data_i;    // Input data from  MMU
  input data_rdy_i;           // Signal from MMU
  input [31:0] va_i;          // Virtual address from the E stage
  input [4:0] ls_reg_i;       // Load/Store register address
  input stall_i;              // Input stall signal
  input [31:0] rd_sc_i;       // Input data from SC opcode from ALU
  input uart_txrdy_i;         // Input from UART
  output [31:0] rd_o;         // Output register
  output [4:0] rda_o;         // Output register address
  output [31:0] mmu_data_o;   // Output data to MMU
  output [31:0] va_o;         // Virtual address to the MMU
  output rtype_o;             // Flag that instruction is of R-type
  output [31:0] rdl_o;        // Register loaded from memory
  output [4:0] rdla_o;        // Address of the register loaded from memory
  output stall_o;             // Output signal to stall I and E stages
  output [5:0] op_o;          // Output opcode
  output load_o;              // Output flag that instruction is Load
  output [1:0] offset01_o;    // Bits [1:0] of offset field
  output [3:0] byte_mask_o;   // Byte mask for MMU
  output uart_wr_o;           // Output to UART
  output [7:0] uart_data_o;   // Output data to UART
  
  reg [31:0] rd_o;
  reg [4:0] rda_o;
  reg rtype_o;
  reg [31:0] rdl_o;
  reg [4:0] rdla_o;
  reg stall_o;
  reg [5:0] op_o;
  reg load_o;
  reg [1:0] offset01_o;
  reg [3:0] byte_mask_o;
  reg [31:0] mmu_data_o;
  
  reg [31:0] rd_buf;          // Buffer to delay rd for 0.5 clk
  reg [4:0] rda_buf;          // Buffer to delay rda for 0.5 clk
  reg [4:0] ls_reg_buf;       // Buffer to delay ls_reg for 0.5 clk
  reg rtype_buf;              // Buffer to delay rtype for 0.5 clk
  reg [31:0] rdl_buf;         // Buffer to delay for 0.5 clk
  reg [5:0] op_buf;           // Buffer to delay for 0.5 clk
  reg load_buf;               // Buffer to delay for 0.5 clk
  reg [1:0] offset01_buf;     // Buffer to delay for 0.5 clk  
  reg tmp_reg; 
  
  reg [3:0] byte_mask_buf;    // Buffer to delay for 0.5 clk
  reg [31:0] mmu_data_buf;    // Buffer to delay for 0.5 clk
  
  // Test logic ----------------------------------------------
  reg [3:0] sw_r = 4'h0;
  reg [3:0] strap_r = 4'h0;
  reg [9:0] count_r = 10'h41;
 
    test_m uart_link (
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .sdata1_i             (mmu_data_buf),
    .sdata2_i             ({op_buf[5:0],26'h0}),
    .sw_i                 (sw_r),
    .strap_i              (strap_r),
    .count_i              (count_r),
    .uart_txrdy_i         (uart_txrdy_i),
    .start_i              (rst_i),
    .uart_wr_o            (uart_wr_o),
    .uart_data_o          (uart_data_o)
  );                    
  

	
  //-------------------------------------------------------------
  
  assign va_o = {va_i[31:2],2'b00};     //Address needs to be corrected in case it's not aligned 
  

 
  always @ (posedge clk_i)
    begin
      case ({op_i,va_i[1:0]})            // Synthesis full case
        {`OP_SB,2'b00}:
          begin
            byte_mask_buf <= 4'b0001;
            mmu_data_buf <= rd_i;
          end  
        {`OP_SB,2'b01}:
          begin
            byte_mask_buf <= 4'b0010;
            mmu_data_buf <= rd_i;
          end
        {`OP_SB,2'b10}:
          begin
            byte_mask_buf <= 4'b0100;
            mmu_data_buf <= rd_i;
          end
        {`OP_SB,2'b11}:
          begin
            byte_mask_buf <= 4'b1000;
            mmu_data_buf <= rd_i;
          end
        {`OP_SH,2'b00}:
          begin
            byte_mask_buf <= 4'b0011;
            mmu_data_buf <= rd_i;
          end
        {`OP_SH,2'b10}: 
          begin
            byte_mask_buf <= 4'b1100;
            mmu_data_buf <= rd_i;
          end
        {`OP_SW,2'b00}: 
          begin
            byte_mask_buf <= 4'b1111;
            mmu_data_buf <= rd_i;
          end
        {`OP_SWL,2'b00}:
           begin
             byte_mask_buf <= 4'b0001;
             mmu_data_buf <= {rd_i[23:0],rd_i[31:24]};   // Only MSB is required - rd_i[31:24]
           end
        {`OP_SWL,2'b01}:
          begin
            byte_mask_buf <= 4'b0011;
            mmu_data_buf <= {rd_i[15:0],rd_i[31:16]};  
          end
        {`OP_SWL,2'b10}: 
          begin
            byte_mask_buf <= 4'b0111; 
            mmu_data_buf <= {rd_i[7:0],rd_i[31:8]};
          end
        {`OP_SWL,2'b11}:
          begin
            byte_mask_buf <= 4'b1111; 
            mmu_data_buf <= rd_i;
          end
        {`OP_SWR,2'b00}:
          begin
            byte_mask_buf <= 4'b1111;
            mmu_data_buf <= rd_i;
          end
        {`OP_SWR,2'b01}: 
          begin
            byte_mask_buf <= 4'b1110;
            mmu_data_buf <= {rd_i[23:0],rd_i[31:24]};
          end
        {`OP_SWR,2'b10}: 
          begin
            byte_mask_buf <= 4'b1100;
            mmu_data_buf <= {rd_i[15:0],rd_i[31:16]};
          end
        {`OP_SWR,2'b11}:
          begin
            byte_mask_buf <= 4'b1000;  
            mmu_data_buf <= {rd_i[7:0],rd_i[31:8]};
          end
        {`OP_SC,2'b00}: 
          begin
            byte_mask_buf <= 4'b1111;
            mmu_data_buf <= rd_sc_i;
          end
      endcase
    end 
   
   always @ (negedge clk_i)
     begin
       mmu_data_o <= mmu_data_buf;
       byte_mask_o <= byte_mask_buf;
     end

  always @ (posedge clk_i)
    begin
      if (!rst_i)   // Reset condition
        begin
          stall_o <= 1'b0;
        end
      else
        begin
          case ({load_i,data_rdy_i})             // Synthesis full case
            2'b11: stall_o <= 1'b0;
            2'b10: stall_o <= 1'b1;             // Data not ready yet, need to wait
            default: stall_o <= 1'b0; 
          endcase
        end 
      
 
      rd_buf <= rd_i;
      rda_buf <= rda_i;   
      rtype_buf <= rtype_i;
      ls_reg_buf <= ls_reg_i;
      op_buf <= op_i;
      load_buf <= load_i;
      offset01_buf <= va_i[1:0];
    end
    
  always @ (negedge clk_i)
    begin
      if (!stall_i)
        begin
          rd_o <= rd_buf;                     // 0.5 clk delay
          rda_o <= rda_buf;                   // 0.5 clk delay
          rtype_o <= rtype_buf;               // 0.5 clk delay
          rdla_o <= ls_reg_buf;               // 0.5 clk delay
          rdl_o <= mmu_data_i;                // 0.5 clk delay
          op_o <= op_buf;                     // 0.5 clk delay
          load_o <= load_buf;                 // 0.5 clk delay
          offset01_o <= offset01_buf;         // 0.5 clk delay
        end
      else
        begin  // If stall happened in the M stage then A stage needs to be slipped -
               // NOP operation to be issued, basically all output signals zero
          rd_o <= 32'h00000000;
          rda_o <= 5'b00000;
          rtype_o <= 1'b0;
          rdla_o <= 5'b00000;
          rdl_o <= 32'h00000000;                 
          op_o <= 6'b000000;
          load_o <= 1'b0;
        end
    end
      
endmodule

