//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
`include "uart_parameters.v"
//
module rx_cntr (
  clk_i, baud_clk_i, rst_i, ffull_i, sin_i,
  data_o, rx_overrun_o, wr_o, frame_err_o
  );
  
  input clk_i;                    // Input clock signal
  input baud_clk_i;               // Input baud clock
  input rst_i;                    // Input reset signal
  input ffull_i;                  // Rx FIFO full
  input sin_i;                    // Serial in
  output [7:0] data_o;            // Data output
  output rx_overrun_o;            // Overrun flag - there was no space in Rx FIFO
  output wr_o;                    // Signal to write data to the Rx FIFO
  output frame_err_o;             // Frame error - if stop bit is not received
  
  reg [7:0] data_o;
  reg wr_o;
  reg frame_err_o;
  
  // Internal registers and wires
  reg [7:0] rx_shiftreg;          // Receiver shift register
  reg [3:0] rx_state; 
  reg [3:0] rx_next_state;
  reg [3:0] baud_counter;
  reg [3:0] bit_counter;
  reg start_rx = 1'b0;                   // Identification of start bit reception
  reg rx_overrun;
  reg stop_bit;                   // Holds value of stop bit
  reg rx_byte_done;               // Flag that reception of one byte is completed
  reg wr_buf; 
  
  assign rx_overrun_o =  rx_overrun;
  
  always @ (rx_state or start_rx or ffull_i or bit_counter or baud_counter or baud_clk_i or sin_i)
    begin
      rx_next_state = `RX_IDLE;
      rx_overrun = 1'b0;
      wr_buf = 1'b0;
      frame_err_o = 1'b0;    
      case (rx_state)                   // Synthesis full case
        `RX_IDLE:
          begin
            if (start_rx && ~ffull_i)
              rx_next_state = `RX_DATA;
            else if (start_rx && ffull_i)
              rx_overrun = 1'b1;
          end
        `RX_DATA:
          begin
            if ((bit_counter === 4'h9) && (baud_counter === 4'h8) && (baud_clk_i === 1'b0))  // baud_clk_i = 1'b0 is made to make sure sampling takes place after negedge
              begin 
                rx_next_state = `RX_IDLE;
                if (sin_i)
                  begin
                    wr_buf = 1'b1;
                    frame_err_o = 1'b0;  
                  end
                else
                  frame_err_o = 1'b1;
              end
            else
              rx_next_state = `RX_DATA;
          end
      endcase
    end
    
    always @ (posedge clk_i)
      begin
        if (!rst_i)
          begin
            rx_state <= `RX_IDLE;
          end
        else
          begin
            rx_state <= rx_next_state;
          end
      end
      
    always @ (negedge clk_i)
      begin
        if (!rst_i)
          begin
            wr_o <= 1'b0;
          end
        else
          begin
            wr_o <= wr_buf;
          end
      end
      
    always @ (posedge baud_clk_i)
      begin
        if ((rx_state == `RX_IDLE) && (sin_i == 1'b1))      // Keep counters zero until sin_i is high and rx_state is IDLE
          begin
            baud_counter <= 4'h0;
            bit_counter <= 4'h0;
          end
        else
          begin
            baud_counter <= baud_counter + 1'b1;
          end
        if (baud_counter === 4'hF)
          bit_counter <= bit_counter + 1'b1;
      end
      
    always @ (negedge baud_clk_i)
      begin
        if ((rx_state == `RX_IDLE) && (baud_counter == 4'h8))
          start_rx <= 1'b1;
        else
          start_rx <= 1'b0;
        if ((rx_state == `RX_DATA) && (baud_counter == 4'h8))
          rx_shiftreg[7:0] <= {sin_i,rx_shiftreg[7:1]};
          
        if ((bit_counter === 4'h9) && (baud_counter === 4'h0))
          data_o <= rx_shiftreg;                                // Store shifted byte
      end
  
endmodule
