//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
module sopc (
  clk_i, dclkp_i, dclkn_i, sin_i, btn_i, sw_i,
  sout_o, led_o
  );
  input clk_i;                    // Input single clock
  input dclkp_i;                  // Positive line of input differential clock
  input dclkn_i;                  // Negative line of input differential clock
  input sin_i;                    // Serial input
  input [3:0] btn_i;              // Input buttons
  input [3:0] sw_i;               // Input GPIO swithes
  output sout_o;                  // Serial output
  output [3:0] led_o;             // Drivers to the LEDs
  
  // Internal registers and wires
  reg cs_r = 1'b1;                // UART always selected
				
  reg [2:0] adr_r = 3'b000;       // Address is constant
	   
  reg [31:0] counter;
  wire rst_low = ~btn_i[0];
 
  wire uart_rxrdy_w;				
  wire uart_txrdy_w;
  wire uart_rd_w;
  wire uart_wr_w;
  wire [7:0] uart_datao_w;
  wire [7:0] uart_datai_w;
	   
  reg [5:0] irq_r = 6'b000000;	
 
  reg irq_btn;                          // Test register
  reg [19:0] btn_counter;                // Test register
 
  
  wire eb_avalid_w;
  wire eb_instr_w;
  wire eb_write_w;
  wire eb_burst_w;
  wire eb_bfirst_w;
  wire eb_blast_w;
  wire [1:0] eb_blen_w;
  wire [3:0] eb_be_w;
	
  wire [31:0] eb_adr_w;
  wire [31:0] eb_datai_w;
  wire eb_ardy_w;
  wire [31:0] eb_datao_w;                  
  wire eb_rdval_w;
  wire eb_wdrdy_w;
  wire eb_rberr_w;
  wire eb_wberr_w;
  wire clk20_w;
  wire clk10_w;		 
  
  wire uart_wro_mips_w;
  wire uart_wro_memctrl_w;
  wire [7:0] uart_datao_mips_w;
  wire [7:0] uart_datao_memctrl_w;
  
  // SW positions
  // 4'b0000 - UART data from MIPS
  // 4'b0001 - UART data from memctrl
  assign uart_wr_w = (sw_i == 4'h0) ? uart_wro_mips_w :
                     (sw_i == 4'h1) ? uart_wro_memctrl_w: 1'b0;
  assign uart_datai_w = (sw_i == 4'h0) ? uart_datao_mips_w :
                        (sw_i == 4'h1) ? uart_datao_memctrl_w: 8'h0;
					
  reg eb_sblock_reg = 1'b0;        // Only sequential sub block ordering is used
  reg init_done_r;
  
  reg rst_mips_pos;
  reg rst_mips;                    // 0.5 clk shifted signal ref to the rst_mips_pos
						   
  
  // Instantiate submodules
   																			  
  clk_wiz_v1_7_0 diff_clk (
     .CLK_IN1_P        (dclkp_i),
     .CLK_IN1_N        (dclkn_i),
     .CLK_OUT1         (clk20_w),                // 20 MHz output
     .CLK_OUT2         (clk10_w)                 // 10 MHz output
  );

  mips_top mips_top_sopc (  
     .clk_i            (clk20_w), 
     .cold_rst_i       (rst_mips), 
     .nmi_i            (), 
     .warm_rst_i       (),
     .int_i            ({irq_r[5:1],irq_btn}), 
     .mergemode_i      (),
     .simplebe_i       (), 
     .eb_ardy_i        (eb_ardy_w), 
     .eb_data_i        (eb_datao_w), 
     .eb_rdval_i       (eb_rdval_w), 
     .eb_wdrdy_i       (eb_wdrdy_w),
     .eb_rberr_i       (eb_rberr_w), 
     .eb_wberr_i       (eb_wberr_w), 
     .eb_ewbe_i        (), 
     .eb_sblock_i      (eb_sblock_reg),	 
     .uart_txrdy_i     (uart_txrdy_w),
     .clk_o            (), 
     .erl_o            (), 
     .exl_o            (), 
     .rp_o             (), 
     .sleep_o          (), 
     .timerint_o       (), 
     .eb_avalid_o      (eb_avalid_w),
     .eb_instr_o       (eb_instr_w), 
     .eb_write_o       (eb_write_w), 
     .eb_burst_o       (eb_burst_w), 
     .eb_bfirst_o      (eb_bfirst_w), 
     .eb_blast_o       (eb_blast_w),
     .eb_blen_o        (eb_blen_w), 
     .eb_be_o          (eb_be_w), 
     .eb_adr_o         (eb_adr_w), 
     .eb_data_o        (eb_datai_w), 
     .eb_wwbe_o        (),
     .uart_wr_o        (uart_wro_mips_w),
     .uart_data_o      (uart_datao_mips_w)
     
  );
 
  mem_ctrl mem_ctrl_sopc (
						
     .clk_i            (clk20_w), 
     .rst_i            (rst_low), 
     .eb_avalid_i      (eb_avalid_w), 
     .eb_instr_i       (eb_instr_w), 
     .eb_write_i       (eb_write_w),
     .eb_burst_i       (eb_burst_w), 
     .eb_bfirst_i      (eb_bfirst_w), 
     .eb_blast_i       (eb_blast_w), 
     .eb_blen_i        (eb_blen_w),
     .eb_be_i          (eb_be_w), 
     .eb_adr_i         (eb_adr_w), 
     .eb_data_i        (eb_datai_w),
     .uart_txrdy_i     (uart_txrdy_w),
     .uart_rxrdy_i     (uart_rxrdy_w),
     .uart_data_i      (uart_datao_w),
     .init_done_i      (init_done_r),
     .sw_i             (sw_i),
     .btn_i            (btn_i),
     .eb_ardy_o        (eb_ardy_w), 
     .eb_data_o        (eb_datao_w), 
     .eb_rdval_o       (eb_rdval_w), 
     .eb_wdrdy_o       (eb_wdrdy_w), 
     .eb_rberr_o       (eb_rberr_w),
     .eb_wberr_o       (eb_wberr_w),
     .uart_rd_o        (uart_rd_w),
     .uart_wr_o        (uart_wro_memctrl_w),
     .uart_data_o      (uart_datao_memctrl_w),
     .led_o            (led_o)
   );

uart_top uart_sopc (  
     .clk_i            (clk20_w), 
     .rst_i            (rst_low),
     .sin_i            (sin_i), 
     .ctsn_i           (),  
     .sout_o           (sout_o), 
     .rtsn_o           (),
     .data_i           (uart_datai_w), 
     .rd_i             (uart_rd_w), 
     .wr_i             (uart_wr_w),
     .adr_i            (adr_r), 
     .cs_i             (cs_r),  
     .data_o           (uart_datao_w), 
     .irq_o            (), 
     .txrdy_o          (uart_txrdy_w), 
     .rxrdy_o          (uart_rxrdy_w)
  );

	   
  always @ (posedge clk_i)
    begin
      if (!rst_low)
        begin
          init_done_r <= 1'b0;
 
          rst_mips_pos <= 1'b0;
        end
			   
      else
        begin
          if (btn_i[1])
            begin
              init_done_r <= 1'b1;
              rst_mips_pos <= 1'b1;
            end
        end 
    end					  
    
  always @ (negedge clk_i)
    begin 
      if (!rst_low)
        begin
          irq_btn <= 1'b0;				   
          btn_counter <= 20'h0;
        end
       else
         begin			 
           if (btn_i[2] || (btn_counter != 20'h0))
             btn_counter <= btn_counter + 1'b1;	 
                         
           if (btn_counter == 20'h10000)
             irq_btn <= 1'b1;
           else
             irq_btn <= 1'b0;				             			                
         end
      rst_mips <= rst_mips_pos;
    end
  endmodule
  