//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// System coprocessor
//
`include "opcodes_def.v"
`include "parms.v"
//
module sys_cp (
  clk_i, cold_rst_i, si_simplebe_i, mergemode_i, eb_sblock_i, int_i,
  sw_int_i, instrio_i, sys_reg_i, va_ex_i, ioverflow_i, set_load_i, set_store_i,
  branch_i, pc_i, cmd_ready_i, pc_next_i, eb_rberr_i, eb_wberr_i, data_access_i,
  access_ex_i, va_i, kernel_mode_i,
  sys_reg_o, pc_o, ex_branch_o, sys_eslip_o, irq_o, irq_branch_o, r12_status_o,
  r23_debug_o, r16_config_o, r16_config1_o
  );
  input clk_i;                        // Input clock
  input cold_rst_i;                   // Input reset signal
  input [1:0] si_simplebe_i;          // Byte enable status, these are static input signal to the MIPS core
  input [1:0] mergemode_i;            // Merge mode - static inputs to the MIPS core
  input eb_sblock_i;                  // Burst order - static input to the MIPS core
  input [5:0] int_i;                  // Interrupt inputs
  input [1:0] sw_int_i;               // Software interrupt inputs
  input [31:0] instrio_i;             // Fetched instruction - output from the I-stage
  input [31:0] sys_reg_i;             // System register input - from MTC0 instruction
  input va_ex_i;                      // Input from E stage to signal load/store exception
  input ioverflow_i;                  // Arithmetic instruction overflow from the E stage exception
  input set_load_i;                   // Flag that va_ex_i is from LOAD instruction
  input set_store_i;                  // Flag that va_ex_i is from STORE instruction
  input branch_i;                     // To make a decision if an instruction is in the branch delay slot
  input [31:0] pc_i;                  // Fetch address from the I stage
  input cmd_ready_i;                  // Input signal from MMU - to find out when instructions fetched
  input [31:0] pc_next_i;             // Next fetch instruction address from the I stage - to monitor alignment
  input eb_rberr_i;                   // Bus error during read operation from BIU
  input eb_wberr_i;                   // Bus error during write operation from BIU
  input data_access_i;                // Signal from MMU that data access is ongoing
  input access_ex_i;                  // Signal from ALU that access violation occured
  input [31:0] va_i;                  // Virtual address - for data exception handling
  input kernel_mode_i;                // Signal from MMU. When = 1 program is in kernel mode
  output [31:0] sys_reg_o;            // System register output - for MFC0 instruction
  output [31:0] pc_o;                 // Output new program counter to the I stage
  output ex_branch_o;                 // Signal to the I stage that exception branch to be taken
  output sys_eslip_o;                 // Output signal to slip E stage
  output irq_o;                       // Signal to the I stage that exception was external interrupt
  output irq_branch_o;                // Signal that IRQ has occured for instruction in the branch delay slot
  output [31:0] r12_status_o;         // Status register for MMU
  output [31:0] r23_debug_o;          // Debug register for MMU
  output [31:0] r16_config_o;         // Config register for MMU
  output [31:0] r16_config1_o;        // Config1 register for MMU
  
  reg [31:0] sys_reg_o;
  reg [31:0] pc_o;
  reg ioverflow_buf;                  // To shift transition of ioverflow from posedge to the negedge
  reg va_ex_load_buf;                 // To shift transition of va_ex_load from posedge to the negedge
  reg va_ex_store_buf;                // To shift transition of va_ex_store from posedge to the negedge
  reg a_eret_buf;                     // To shift transition of a_eret from negedge to the posedge
  reg a_syscall_buf;                  // To shift transition of a_syscall to the posedge
  reg a_syscall_bufn;                 // Second buffer for syscall - to delay instruction for one cycle
  reg a_reserved_buf;                 // To shift transition of a_reserved to the posedge
  reg a_reserved_bufn;                // Second buffer for a_reserved - to delay instruction for one cycle
  reg a_cpu_buf;                      // To shift transition of a_cpu to the posedge
  reg a_cpu_bufn;                     // Second buffer for a_cpu - to delay instruction for one cycle
  reg a_not_implemented_buf;          // To shift transition of a_not_implemented to the posedge
  reg a_not_implemented_bufn;         // Second buffer for a_not_implemented - to delay instruction for one cycle
  reg branch_buf1;                    // Two branch buffers requred to delay branch pulse for one clock
  reg branch_buf2;       
  reg irq_in_buf;                     // Buffer to make sure irq is registered on the posedge
  reg irq_in_bufn;                    // 0.5 clk delay of irq_in_buf - registered on the negedge
  reg cmd_readyi_neg;                 // Two registers to shift cmd_ready_i one clock ahead
  reg cmd_readyi_pos;                 // Second buffer to make one clock delay
  reg branch_status;                  // Flag to find out if current instruction causes branch
  reg user_fetch_ex;                  // Fetch exception - when user program tries to access kernel segment
  
  assign irq_o = irq_in_buf;
  assign irq_branch_o = branch_buf1;
   
  // Internal wires and registers
  reg [31:0] r8_badvaddr;   // Read Only register to capture the most recent virtual address
                            // that caused address error (AdEL or AdES)
  reg [31:0] r9_count;      // Interval counter, incremented every other clock, R/W
  reg [31:0] r11_compare;   // R/W register used in conjunction with count register
                            // Generates SI_TimerInt output interrupt and normally fed back
                            // as HW interrupt 5
  reg [31:0] r12_status;    // Status register. Some fields are Read Only while the other are R/W
  reg [31:0] r13_cause;     // Cause of the most recent exception, primarily is Read Only with a few exceptions
  reg [31:0] r14_epc;       // Exception Program Counter, R/W
  reg [31:0] r15_prid = 32'h00118301;     // Processor identification, Read Only, revision set to 1.1-2.2 
  reg [31:0] r16_config;    // Config register, Read Only and R/W
  reg [31:0] r16_config1;   // Adjucent to config register. Read Only
  reg [31:0] r17_lladdr;    // Physical address read by the most recent LL instruction, Read Only
  reg [31:0] r18_watchlo;   // WatchLo register, R/W
  reg [31:0] r19_watchhi;   // WatchHi register, R/W
  reg [31:0] r23_debug;     // Debug register, Read Only and R/W
  reg [31:0] r24_depc;      // Debug exception program counter, R/W
  reg [31:0] r26_errctl;    // Error control register, R/W
  reg [31:0] r28_taglo;     // TagLo register, R/W
  reg [31:0] r28_datalo;    // DataLo register, R/W
  reg [31:0] r30_errorepc;  // ErrorEPC register, R/W
  reg [31:0] r31_desave;    // Debug Exception Save register, R/W
  
  assign r12_status_o = r12_status;
  assign r23_debug_o = r23_debug;
  assign r16_config_o = r16_config;
  assign r16_config1_o = r16_config1;
  
  reg count_inc;      // Register to count other cycles for r9_count register
  reg irq_ex;         // When irq_ex = 1'b1 interrupt occured during exception handling
  reg va_ex_fetch;    // Register to monitor alignment of PC
  reg va_ex_fetch_bufn;   // To shift transition of va_ex_fetch to the negedge
  reg bus_error_bufn;     // To shift transition of bus_error to the negedge
  reg access_exi_bufn;    // To shift transition of access_ex_i to the negedge
  reg ex_brancho_bufn;    // To shift transition of ex_branch_o to the negedge
  reg kernel_modei_bufn;  // To shift transition of kernel_mode_i to the negedge
  reg user_fetch_ex_bufn; // To shift transition of user_fetch_ex to the negedge
  
  
  wire a_mfc0 = (instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == `CZ_MFC0);
  wire a_mtc0 = (instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == `CZ_MTC0); 
  wire a_eret = (instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == `CO_ERET);
  wire a_syscall = (instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == `SP_SYSCALL);
  wire a_pref = (instrio_i[31:26] == `OP_PREF);       // CP0 does nothing on this instruction
  wire a_cache = (instrio_i[31:26] == `OP_CACHE);     // CP0 does nothing on this instruction
  wire a_reserved = (instrio_i[31:28] == 4'b0110) ||
                    (instrio_i[31:26] == 6'b011101) ||
                    (instrio_i[31:27] == 5'b01111) ||
                    (instrio_i[31:26] == 6'b100111) ||
                    (instrio_i[31:27] == 5'b10110) ||
                    (instrio_i[31:26] == 6'b110100) ||
                    (instrio_i[31:26] == 6'b110111) ||
                    (instrio_i[31:26] == 6'b111011) ||
                    (instrio_i[31:26] == 6'b111100) ||
                    (instrio_i[31:26] == 6'b111111) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b000101)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b001110)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:2] == 4'b0101)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:2] == 4'b0111)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:1] == 5'b10100)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:2] == 4'b1011)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b110101)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b110111)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:3] == 3'b111)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:0] == 6'b000011)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:1] == 5'b00011)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:3] == 3'b011)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:4] == 2'b01)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:1] == 5'b10001)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:2] == 4'b1001)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:3] == 3'b101)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:3] == 3'b110)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:2] == 4'b1110)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:1] == 5'b11110)) ||
                    ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:0] == 6'b111110)) ||
                    ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:18] == 3'b001)) ||
                    ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:16] == 5'b01101)) ||
                    ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:16] == 5'b01111)) ||
                    ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:18] == 3'b101)) ||
                    ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:19] == 2'b11)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b00001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:22] == 4'b0001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b00101)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:22] == 4'b0011)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:23] == 3'b010)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:22] == 4'b0110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:4] == 2'b00)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:3] == 3'b010)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == 6'b011001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:1] == 5'b01101)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:1] == 5'b01110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == 6'b011110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == 6'b100001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:1] == 5'b10001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:2] == 4'b1001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:3] == 3'b101)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:4] == 2'b11)) ||                    
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:4] == 2'b00)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:3] == 3'b010)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:0] == 6'b011001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:1] == 5'b01101)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:1] == 5'b01110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:0] == 6'b011110)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:0] == 6'b100001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:1] == 5'b10001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:2] == 4'b1001)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:3] == 3'b101)) ||
                    ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:4] == 2'b11));

wire a_cpu =  (instrio_i[31:26] == 6'b010001) ||
              (instrio_i[31:27] == 5'b01001) ||
              (instrio_i[31:26] == 6'b110001) ||
              (instrio_i[31:26] == 6'b110010) ||
              (instrio_i[31:26] == 6'b110101) ||
              (instrio_i[31:26] == 6'b110110) ||
              (instrio_i[31:26] == 6'b111001) ||
              (instrio_i[31:26] == 6'b111010) ||
              (instrio_i[31:26] == 6'b111101) ||
              (instrio_i[31:26] == 6'b111110) ||
              ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b000001));
              
wire a_not_implemented =  ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b001101)) ||         // BREAK opcode
                          ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b001111)) ||         // SYNC opcode
                          ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:2] == 4'b1100)) ||           // TGE, TGEU, TLT, TLTU opcodes
                          ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b110100)) ||          // TEQ opcode
                          ((instrio_i[31:26] == `OP_SPECIAL) && (instrio_i[5:0] == 6'b110110)) ||          // TNE opcode
                          ((instrio_i[31:26] == `OP_SPECIAL2) && (instrio_i[5:0] == 6'b111111)) ||         // SDBBP opcode
                          ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:18] == 3'b010)) ||            // TGEI, TGEIU, TLTI, TLTIU opcodes
                          ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:16] == 5'b01100))  ||         // TEQI opcode
                          ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:16] == 5'b01110)) ||          // TNEI opcode
                          ((instrio_i[31:26] == `OP_REGIMM) && (instrio_i[20:17] == 4'b1001)) ||           // BLTZALL, BGEZALL opcodes
                          ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == 6'b011111)) ||   // DERET opcode
                          ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25] == 1'b1) && (instrio_i[5:0] == 6'b100000)) ||   // WAIT opcode
                          ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:0] == 6'b011111)) ||   // DERET opcode
                          ((instrio_i[31:26] == `OP_COP0) && (instrio_i[25:21] == 5'b01111) && (instrio_i[5:0] == 6'b100000));   // WAIT opcode
                          

  // Wires for MFC0 instruction
  wire [31:0] sys_reg_wire = ({instrio_i[15:11],instrio_i[2:0]} === 8'b01000000) ? r8_badvaddr : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01001000) ? r9_count : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01011000) ? r11_compare : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01100000) ? r12_status : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01101000) ? r13_cause :
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01110000) ? r14_epc : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b01111000) ? r15_prid : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10000000) ? r16_config : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10000001) ? r16_config1 : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10001000) ? r17_lladdr : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10010000) ? r18_watchlo : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10011000) ? r19_watchhi : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b10111000) ? r23_debug : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11000000) ? r24_depc : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11010000) ? r26_errctl : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11100000) ? r28_taglo : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11100001) ? r28_datalo : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11110000) ? r30_errorepc : 
                             ({instrio_i[15:11],instrio_i[2:0]} === 8'b11111000) ? r31_desave : 
                             32'hzzzzzzzz;

  // Wires to mask/unmask interrupts
  wire hw_int5 = int_i[5] & r12_status[15];   // r12 contains masks for the interrupts
  wire hw_int4 = int_i[4] & r12_status[14];
  wire hw_int3 = int_i[3] & r12_status[13];
  wire hw_int2 = int_i[2] & r12_status[12];
  wire hw_int1 = int_i[1] & r12_status[11];
  wire hw_int0 = int_i[0] & r12_status[10];
  wire sw_int1 = sw_int_i[1] & r12_status[9];
  wire sw_int0 = sw_int_i[0] & r12_status[8];
  
  wire va_ex_load = va_ex_i & set_load_i;       // To distinct that va_ex_i occured for load instruction
  wire va_ex_store = va_ex_i & set_store_i;     // To distinct that va_ex_i occured for store instruction
  wire bus_error = eb_rberr_i | eb_wberr_i;     // Bus error for read and write operations
  
  // Interrupts enabled only if IE=1 && EXL=0 && ERL=0 && DM=0
  // Interrupt waits until data access is completed (~data_access_i)
  wire irq_in = (r13_cause[15] | r13_cause[14] | r13_cause[13] | r13_cause[12] | r13_cause[11] | 
                 r13_cause[10] | r13_cause[9] | r13_cause[8]) 
                && r12_status[0] && ~r12_status[1] && ~r12_status[2] && ~r23_debug[0] && ~data_access_i;
  
  // Wire to signal if exception condition has occured - must be changing on negedge
  wire ex_in = va_ex_load_buf || va_ex_store_buf || ioverflow_buf || irq_in_bufn || a_syscall_bufn || a_reserved_bufn || a_cpu_bufn
                || a_not_implemented_bufn || va_ex_fetch_bufn || bus_error_bufn || access_exi_bufn ||
                 user_fetch_ex_bufn;

  // All signals in ex_branch_o must be changing on the posedge
  assign ex_branch_o = ioverflow_i || va_ex_load || va_ex_store || a_eret_buf || irq_in_buf || a_syscall_buf || a_reserved_buf || a_cpu_buf 
                || a_not_implemented_buf || va_ex_fetch || bus_error || access_ex_i || user_fetch_ex;

  assign sys_eslip_o = irq_in_buf;      // This is required to cancel interrupted operation
  
  always @ (a_mfc0)
    begin
      if (a_mfc0)
        sys_reg_o = sys_reg_wire;
    end
  
  always @ (*)                        // Logic for exception program counter
    begin                             // Control signals (ioverflow_i, a_eret_buf) must be changing on posedge
      if (a_eret_buf)
        begin
          if (r12_status[2])          // Check if ERL bit was set
            pc_o = r30_errorepc;
          else
            pc_o = r14_epc;
        end
      else if (irq_in_buf)
        case ({r12_status[22],r13_cause[23]})     // STATUSbev and CAUSEiv bits
          2'b00: pc_o = `IRQ0;                   
          2'b01: pc_o = `IRQ1;
          2'b10: pc_o = `IRQ2;
          2'b11: pc_o = `IRQ3;
        endcase      
      else if ((ioverflow_i || va_ex_load || va_ex_store || a_syscall_buf || a_reserved_buf ||
                a_cpu_buf || a_not_implemented_buf || va_ex_fetch || bus_error ||
                access_ex_i || user_fetch_ex) && ~r12_status[1])
        begin
          if (r12_status[22])
            pc_o = `EX_OTHER1;
          else
            pc_o = `EX_OTHER0;
        end
    end
  
  always @ (posedge clk_i)
    begin
      if (!cold_rst_i)                              // Reset condition
        begin
          r9_count <= 32'h0;                        // Even though specified as undefined in docs
          r12_status <= 32'b00000000010000001000000000000001;   // UM=0,ERL=0,EXL=0,IE=0 (kernel mode)  
          r13_cause <= 32'h0;
          r14_epc <= 32'h00118301;
          r16_config <= {10'b1010011000,si_simplebe_i[0],2'b10,mergemode_i[1:0],eb_sblock_i,16'b0000000110000010};
          r16_config1 <= 32'b00000000100110110100110110001010;
          r17_lladdr <= 32'h0;
          r18_watchlo <= 32'h0;
          r19_watchhi <= 32'h0;
          r23_debug <= 32'b00000010000000010000000000000000;
          r26_errctl <= 32'h0;
          r28_taglo <= 32'h0;
          va_ex_fetch <= 1'b0;
          user_fetch_ex <= 1'b0;
        end
      else
        begin
          case (1'b1)                       // Synthesis full case
            a_mtc0:                       // MTC0 opcode processing
              case ({instrio_i[15:11],instrio_i[2:0]})              // Synthesis full case
                8'b01001000: r9_count <= sys_reg_i;
                8'b01011000: r11_compare <= sys_reg_i;
                8'b01100000: r12_status <= sys_reg_i;
                8'b01101000: r13_cause <= sys_reg_i;
                8'b01110000: r14_epc <= sys_reg_i;
                8'b10000000: r16_config <= sys_reg_i;
                8'b10010000: r18_watchlo <= sys_reg_i;
                8'b10011000: r19_watchhi <= sys_reg_i;
                8'b10111000: r23_debug <= sys_reg_i;
                8'b11000000: r24_depc <= sys_reg_i;
                8'b11010000: r26_errctl <= sys_reg_i;
                8'b11100000: r28_taglo <= sys_reg_i;
                8'b11100001: r28_datalo <= sys_reg_i;
                8'b11110000: r30_errorepc <= sys_reg_i;
                8'b11111000: r31_desave <= sys_reg_i;
              endcase
            a_eret:                         // ERET opcode processing
              begin
                if (r12_status[2])          // Check if ERL bit was set
                  r12_status[2] <= 1'b0;      // Clear ERL bit
                else
                  r12_status[1] <= 1'b0;      // Clear EXL bit
              end              
          endcase
        
          if (ex_in && ~r12_status[1])      // Exception has occured and can be handled - EXL bit is cleared
            begin
              if (branch_buf2)
                begin
                  r14_epc <= pc_i - 4'h4;           // Load restart address
                  r13_cause[31] <= 1'b1;            // BD - branch delay bit
                end
              else
                begin
                  if (a_syscall_bufn)
                    r14_epc <= pc_i + 4'h4;           // For syscall exception need to return to the next instruction
                  else if (!(irq_in_bufn & irq_ex))      // If IRQ is taken from exception r14_epc does not require update
                    r14_epc <= pc_i;
                  r13_cause[31] <= 1'b0;            // Clear BD - branch delay bit
                end
              r13_cause[29:28] <= 2'b00;            // CE bits. Coprocessor unit number - 0
              r12_status[1] <= 1'b1;                // Set EXL bit in the status register
              if (irq_in_bufn)   // Handling of exceptions based on priority levels
                r13_cause[6:2] <= `EX_INT;  // Exception code
              else if (bus_error_bufn)
                r13_cause[6:2] <= `EX_IBE;  // Bus error
              else if (user_fetch_ex)       // Instruction access violation
                r13_cause[6:2] <= `EX_FE;
              else if (access_exi_bufn)     // Data access violation
                begin
                  r13_cause[6:2] <= `EX_ACS;
                  r8_badvaddr <= va_i;
                end              
              else if (ioverflow_buf)
                r13_cause[6:2] <= `EX_OV;     // Exception code
              else if (va_ex_load_buf | va_ex_fetch_bufn)
                begin 
                  r13_cause[6:2] <= `EX_ADEL;
                  r8_badvaddr <= va_i;
                end
              else if (va_ex_store_buf)
                begin
                  r13_cause[6:2] <= `EX_ADES;
                  r8_badvaddr <= va_i;
                end
              else if (a_syscall_bufn)
                r13_cause[6:2] <= `EX_SYS;
              else if (a_reserved_bufn)
                r13_cause[6:2] <= `EX_RI;
              else if (a_cpu_bufn)
                r13_cause[6:2] <= `EX_CPU;
              else if (a_not_implemented_bufn)
                r13_cause[6:2] <= `EX_NI;
            end
          // Interrupt bits in the cause register set regardless from IE bit
          // To be able to poll cause register with IE bit set to 0
          if (hw_int5)
            r13_cause[15] <= 1'b1;
          if (hw_int4)
            r13_cause[14] <= 1'b1;
          if (hw_int3)
            r13_cause[13] <= 1'b1;
          if (hw_int2)
            r13_cause[12] <= 1'b1;
          if (hw_int1)
            r13_cause[11] <= 1'b1;
          if (hw_int0)
            r13_cause[10] <= 1'b1;
          if (sw_int1)
            r13_cause[9] <= 1'b1;
          if (sw_int0)
            r13_cause[8] <= 1'b1;
          
          if (pc_next_i[1:0] !== 2'b00)   // To check for address alignment
            va_ex_fetch <= 1'b1;
          else 
            va_ex_fetch <= 1'b0;
          if (pc_next_i[31] & ~kernel_modei_bufn)       // To check if user is accessing
            begin                                       // kernel segments
              if (~ex_brancho_bufn)
                user_fetch_ex <= 1'b1;
              else
                user_fetch_ex <= 1'b0;
            end
          else
            user_fetch_ex <= 1'b0;
             
            
        end
          a_eret_buf <= a_eret;  
          a_syscall_buf <= a_syscall;
          a_reserved_buf <= a_reserved;
          a_cpu_buf <= a_cpu;
          a_not_implemented_buf <= a_not_implemented;
          branch_buf1 <= branch_status;
          irq_in_buf <= irq_in;     
          cmd_readyi_pos <= cmd_readyi_neg;                    

    end
  
    
  always @ (negedge clk_i)
    begin
      if (!cold_rst_i)
        begin
          branch_status <= 1'b0;
        end
      else
        begin
          if (cmd_readyi_pos)
            branch_status <= branch_i;
        end
      ioverflow_buf <= ioverflow_i;                   // Buffer to shift transition to the negedge     
      va_ex_store_buf <= va_ex_store;         
      va_ex_load_buf <= va_ex_load;
      branch_buf2 <= branch_buf1;
      irq_in_bufn <= irq_in_buf;
      a_syscall_bufn <= a_syscall_buf;
      a_reserved_bufn <= a_reserved_buf;
      a_cpu_bufn <= a_cpu_buf;
      a_not_implemented_bufn <= a_not_implemented_buf;
      cmd_readyi_neg <= cmd_ready_i;
      va_ex_fetch_bufn <= va_ex_fetch;
      bus_error_bufn <= bus_error;
      access_exi_bufn <= access_ex_i;
      ex_brancho_bufn <= ex_branch_o;
      kernel_modei_bufn <= kernel_mode_i;
      user_fetch_ex_bufn <= user_fetch_ex;
    end
   
  always @ (posedge irq_in)
    begin
      if (!cold_rst_i)
        irq_ex <= 1'b0;
      else
        begin
          if (a_eret)
            irq_ex <= 1'b1;
          else
            irq_ex <= 1'b0;
        end
    end
    
     
endmodule
