//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//		 
// Test module to transmit required data via UART to				 
// the host PC														 
// Two sets of data maximum 1024x32 words each set
// Data recorded to memory on the negedge - signals to be 
// recorded must be asserted on the posedge
// Control signals and data to UART asserted on the posedge
//
module test_m (
  clk_i, rst_i, sdata1_i, sdata2_i, sw_i, strap_i, count_i, uart_txrdy_i,	   
  start_i,
  uart_wr_o, uart_data_o
  );
  input clk_i;                    // Input clock
  input rst_i;                    // Input reset signal
  input [31:0] sdata1_i;          // First data input
  input [31:0] sdata2_i;          // Second data input
  input [3:0] sw_i;               // Input switch to select mode of operation
  input [3:0] strap_i;            // Strapping to set address
  input [9:0] count_i;            // Number of samples to be collected	 
  input uart_txrdy_i;             // Flag that UART is ready to accept the data
  input start_i;                  // Signal to start data acquisition
  output uart_wr_o;               // Write enable signal for UART
  output [7:0] uart_data_o;       // Data for UART
  
  reg uart_wr_o;
  reg [7:0] uart_data_o;
  
  // Internal wires and registers
  reg [11:0] uart_counter;
  reg [31:0] test_data1 [0:31];   // First set of test registers
  reg [31:0] test_data2 [0:31];   // Second set of test registers
  reg [31:0] data_hold;           // Hold register to store data to be Tx'd
  reg [1:0] byte_incr;            // Register to count four bytes in 32-bit word
   
  reg [9:0] counti_buf;           // Register to memorize number of words to be acquired
  reg [9:0] countt1_buf;          // The same buffer as counti_buf, but used when first set of data to be Tx'd to UART
  reg [9:0] countt2_buf;          // The same buffer as counti_buf, but used when second set of data to be Tx'd to UART
  reg start_r;                    // Register to store start value
  reg [9:0] data_ptr;             // Address for memories 
  
  reg mem1_we_r;
  reg mem2_we_r; 
  reg clk2_r = 1'b0;
  reg uart_txrdy_neg;            // To make transition of uart_txrdy on the negedge
  
  wire [31:0] mem1_datai_w;
  wire [9:0] mem1_adri_w;
  wire [31:0] mem1_datao_w;
  
  wire [31:0] mem2_datai_w;
  wire [9:0] mem2_adri_w;
  wire [31:0] mem2_datao_w;
  
  
  // Instantiate blocks
  memory_tm mem1 (
    .clk_i            (clk_i),
    .we_i             (mem1_we_r),
    .data_i           (sdata1_i),
    .adr_i            (data_ptr),
    .data_o           (mem1_datao_w)
  );
  memory_tm mem2 (
    .clk_i            (clk_i),
    .we_i             (mem2_we_r),
    .data_i           (sdata2_i),
    .adr_i            (data_ptr),
    .data_o           (mem2_datao_w)
  );
  
  always @ (posedge clk_i)    // Posedge is required as control signals for the UART must be asserted on the posedge
    begin							
      if (sw_i === strap_i)        // Check if the module has been selected as active
        begin
          if (start_r == 1'b0)
            begin
              counti_buf <= count_i; 
              countt1_buf <= count_i;
              countt2_buf <= count_i;
              data_ptr <= 10'b1111111111;
              uart_wr_o <= 1'b0;
              mem1_we_r <= 1'b0;
              mem2_we_r <= 1'b0;
            end
          else
            begin
              if ((counti_buf != 10'h0) && (counti_buf != 10'h1))        // Acquire data 
                begin
                  data_ptr <= data_ptr + 1'b1;
                  mem1_we_r <= 1'b1;
                  mem2_we_r <= 1'b1;
                  counti_buf <= counti_buf - 1'b1;
                end					 
              else if (counti_buf == 10'h1)        // All the data has been acquired. 
                begin
                  data_ptr <= 10'h0;
                  byte_incr <= 2'b00;	  
                  counti_buf <= counti_buf - 1'b1;
                  mem1_we_r <= 1'b0;
                  mem2_we_r <= 1'b0;
                end
              else if ((countt1_buf != 10'h0) && (countt1_buf != 10'h1))
                begin                // All the data has been acquired and needs to be sent to UART
                  if (uart_txrdy_neg && clk2_r)
                    begin
                      case (byte_incr)
                        2'b00: 
                          begin
                            uart_data_o <= mem1_datao_w[31:24];
                            byte_incr <= byte_incr + 1'b1;                          
                          end								
                        2'b01:
                          begin
                            uart_data_o <= mem1_datao_w[23:16];
                            byte_incr <= byte_incr + 1'b1;
                          end								
                        2'b10:
                          begin
                            uart_data_o <= mem1_datao_w[15:8];
                            byte_incr <= byte_incr + 1'b1;
                          end								
                        2'b11:
                          begin
                            uart_data_o <= mem1_datao_w[7:0];
                            byte_incr <= byte_incr + 1'b1;
                            data_ptr <= data_ptr + 1'b1;
                            countt1_buf <= countt1_buf - 1'b1;
                          end
                      endcase
                      uart_wr_o <= 1'b1;
                    end					
                  else
                    uart_wr_o <= 1'b0;
                end
              else if (countt1_buf == 10'h1)
                begin
                  uart_wr_o <= 1'b0;
                  data_ptr <= 10'h0;
                  byte_incr <= 2'b00;	   
                  countt1_buf <= countt1_buf - 1'b1;
                end						
              else if ((countt2_buf != 10'h0) && (countt2_buf != 10'h1))
                begin
                  if (uart_txrdy_neg && clk2_r)
                    begin
                      case (byte_incr)
                        2'b00:
                          begin
                            uart_data_o <= mem2_datao_w[31:24];
                            byte_incr <= byte_incr + 1'b1;
                          end
                        2'b01:
                          begin
                            uart_data_o <= mem2_datao_w[23:16];
                            byte_incr <= byte_incr + 1'b1;
                          end
                        2'b10:
                          begin
                            uart_data_o <= mem2_datao_w[15:8];
                            byte_incr <= byte_incr + 1'b1;
                          end
                        2'b11:
                          begin
                            uart_data_o <= mem2_datao_w[7:0];
                            byte_incr <= byte_incr + 1'b1;
                            data_ptr <= data_ptr + 1'b1;
                            countt2_buf <= countt2_buf - 1'b1;
                          end
                        endcase
                        uart_wr_o <= 1'b1;
                    end
                  else
                    uart_wr_o <= 1'b0;
                end
              else if (countt2_buf == 10'h1)
                uart_wr_o <= 1'b0;
              else 
                uart_wr_o <= 1'b0;
            end
        end
      else
        begin
          uart_wr_o <= 1'bz;
          uart_data_o <= 8'hzz;
        end   
    end 
    
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          start_r <= 1'b0;
        end
      else
        begin
          if (start_i)
            start_r <= 1'b1;
        end	 
      uart_txrdy_neg <= uart_txrdy_i;
      clk2_r <= ~clk2_r;
    end
    
endmodule