//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
`include "uart_parameters.v"
//
module tx_cntr (
  clk_i, baud_clk_i, rst_i, ftempty_i, data_i,
  sout_o, rd_o
  );
  
  input clk_i;                          // Input clock signal
  input baud_clk_i;                     // Input baud clock
  input rst_i;                          // Input reset signal
  input ftempty_i;                      // Tx FIFO empty
  input [7:0] data_i;                   // Data read from FIFO
  output sout_o;                        // Serial output
  output rd_o;                          // Request to read one word from Tx FIFO
  
  reg rd_o;
  reg sout_o = 1'b1;                    // Line is idle upon power up     
  
  // Internal registers and wires
  reg [7:0] tx_shiftreg;                // Transmitter shift register
  reg [3:0] tx_state;
  reg [3:0] tx_next_state;
  reg rd_buf;                           
  reg [3:0] bus_counter = 4'h0;
  reg [2:0] baud_counter;               // Counter to divide baud clock
  reg baud_16;                          // Baud_clk_i/16 pin
  reg tx_byte_done;                     // Flag that one byte has been Tx'd over the serial line
  
  always @ (tx_state or ftempty_i or bus_counter or tx_byte_done)
    begin
      tx_next_state = `TX_IDLE;
      rd_buf = 1'b0;
      case (tx_state)                       // Synthesis full case
        `TX_IDLE:
          begin
            if (~ftempty_i)                  // If FIFO is not empty go to the data transmission
              begin
                tx_next_state = `TX_DATA;
                rd_buf = 1'b1;
              end
          end
        `TX_DATA:
          begin
            if ((bus_counter == 4'h9) && (tx_byte_done == 1'b1))
              tx_next_state = `TX_IDLE;                             
            else
              tx_next_state = `TX_DATA;
          end
      endcase
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          tx_state <= `TX_IDLE;
        end
      else
        begin
          tx_state <= tx_next_state;
        end
    end
  
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          rd_o <= 1'b0;
          tx_byte_done <= 1'b0;
        end
      else
        begin
          rd_o <= rd_buf;                         // To shift command transition to the negedge
          if (tx_state == `TX_IDLE)
            tx_byte_done <= 1'b0;
          else if ((tx_state == `TX_DATA) && (bus_counter == 4'h8))
            tx_byte_done <= 1'b1;
        end
    end
    
  always @ (posedge baud_clk_i)
    begin
      if (!rst_i)
        begin
          baud_counter <= 3'b000;
          baud_16 <= 1'b0;
        end
      else
        begin
          baud_counter <= baud_counter + 1'b1;
          if (baud_counter === 3'b111)
            baud_16 <= ~baud_16;                        // Baud clock divided by 16
        end
    end
    
  always @ (posedge baud_16)
    begin
      if (!rst_i)
        begin
          bus_counter <= 4'h0;
        end
      else
        begin
          if ((tx_state == `TX_IDLE) || (bus_counter == 4'h9))
            bus_counter <= 4'h0;
          else
            bus_counter <= bus_counter + 1'b1;
        end
    end

  always @ (negedge baud_16)
    begin
      if (tx_state == `TX_DATA)
        begin
          case (bus_counter)                  // Synthesis full case
            4'h0:
              sout_o <= 1'b1;                 // Stop bit
            4'h1: 
              begin
                sout_o <= 1'b0;               // Issue start bit
                tx_shiftreg <= data_i;
              end
            4'h2,4'h3,4'h4,4'h5,4'h6,4'h7,4'h8,4'h9:
              begin
                sout_o <= tx_shiftreg[0];
                tx_shiftreg <= (tx_shiftreg >> 1'b1);
              end
          endcase
        end
      else
        begin
          if (bus_counter == 4'h9)
            begin
              sout_o <= tx_shiftreg[0];   // This line is required to Tx the last bit when state machine is
            end                           // already IDLE
          else
            sout_o <= 1'b1;                 // Idle state
        end
    end  
endmodule