//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// UART top module
// Implements 16550-like UART
//
module uart_top (
  clk_i, rst_i,
  // Serial interface
  sin_i, ctsn_i,  
  sout_o, rtsn_o,
  // CPU interface
  data_i, rd_i, wr_i, adr_i, cs_i,    // Must be changing on the posedge
  data_o, irq_o, txrdy_o, rxrdy_o
  
  );
  
  input clk_i;                    // Input clock signal
  input rst_i;                    // Input reset signal
  input sin_i;                    // Serial data input
  input ctsn_i;                   // Clear to send
  output sout_o;                  // Serial data output
  output rtsn_o;                  // Ready to send
  
  input [7:0] data_i;             // Input data
  input rd_i;                     // Signal from CPU to read UART's registers
  input wr_i;                     // Signal from CPU to write to the UART's registers
  input [2:0] adr_i;              // UART's registers address
  input cs_i;                     // Chip select
  output [7:0] data_o;            // Output data
  output irq_o;                   // Interrupt request
  output txrdy_o;                 // Flag that UART is ready to Tx data
  output rxrdy_o;                 // Flag that UART is ready to Rx data
  
  // Internal registers and wires
  reg [7:0] line_control;         // Line control register - address 011 (DLAB - x)
  reg [7:0] divisor_latch_l;      // Least significant byte of the divisor latch - address 000 (DLAB - 1)
  reg [7:0] divisor_latch_m;      // Most significant byte of the divisor latch - address 001 (DLAB - 1)
  reg [7:0] irq_enable;           // Interrupt enable, address 001 (DLAB - 0)
  reg [7:0] irq_ident;            // Interrupt identification (read); address 010 (DLAB - x)
  reg [7:0] fifo_control;         // FIFO control (write); address 010 (DLAB - x)
  reg [7:0] fifo_th;              // Rx FIFO threshold before issuing interrupt, address 100
  wire [7:0] fifo_status;          // FIFO (both Tx and Rx) status (read), address 101
  reg irq_rxcount;                // Interrupt when there is data in Rx FIFO more than Rx threshold
  reg irq_txfifo_empty;           // Interrupt when Tx FIFO is empty
  
  wire baudout_w;
  wire ftempty_w;
  wire [7:0] txfifo_datao_w;      // I/O directions relative to the tx_fifo
  reg [7:0] txfifo_datai_r;      
  wire txfifo_rdi_w;
  wire frfull_w;
  wire [7:0] rxfifo_datai_w;      // I/O directions relative to the rx_fifo
  wire [7:0] rxfifo_datao_w;
  wire rxfifo_wri_w;
  wire txfifo_full_w;
  wire rxfifo_empty_w;
  wire [8:0] rxfifo_count_w;
  wire rx_overrun_w;
  wire rxframe_err_w;
  reg rx_overrun_neg;             // To make sure transition is on the negedge
  reg rxframe_err_neg;            // To make sure transition is on the negedge
  reg txfifo_wei_r; 
  reg rdrx_neg;                  // To make sure transitions of rd_rx is on the negedge
  reg rd_avail_neg;
  reg rd_avail_pos;              // This signal is delayed one cycle after rd_i and cs_i are asserted
  
  assign txrdy_o = ~txfifo_full_w;
  assign rxrdy_o = ~rxfifo_empty_w;
  
  wire [7:0] data_buf = ((adr_i === 3'b000) && (line_control[7] === 1'b1)) ? divisor_latch_l:
                  ((adr_i === 3'b000) && (line_control[7] === 1'b0)) ? rxfifo_datao_w:
                  ((adr_i === 3'b001) && (line_control[7] === 1'b1)) ? divisor_latch_m:
                  ((adr_i === 3'b001) && (line_control[7] === 1'b0)) ? irq_enable:
                  (adr_i === 3'b010) ? irq_ident:
                  (adr_i === 3'b011) ? line_control:
                  (adr_i === 3'b100) ? fifo_th: 
                  (adr_i === 3'b101) ? fifo_status: 8'hzz;
  assign data_o = (rd_avail_pos) ? data_buf: 8'hzz;
  wire rd_rx = ((adr_i === 3'b000) && (line_control[7] === 1'b0)) ? rd_i : 1'b0;
  // Interrupt enable logic
  // irq_enable[0] - If data in Rx FIFO exceeds threshold
  // irq_enable[1] - If Tx FIFO is empty
  //
  assign irq_o = (irq_rxcount && irq_enable[0]) || (irq_txfifo_empty && irq_enable[1]) ||
                 (rx_overrun_neg && irq_enable[2]) || (rxframe_err_neg && irq_enable[3]);
  
  // Modules instantiation
  baud_generator baud_generator_top (
    .clk_i                (clk_i), 
    .rst_i                (rst_i),
    .divisor_msb_i        (divisor_latch_m), 
    .divisor_lsb_i        (divisor_latch_l), 
    .load_divisor_i       (line_control[7]),
    .baudout_o            (baudout_w)    
  );
  
  fifo tx_fifo_top (
    .clk_i                (clk_i), 
    .rst_i                (rst_i), 
    .we_i                 (txfifo_wei_r), 
    .rd_i                 (txfifo_rdi_w), 
    .data_i               (txfifo_datai_r),
    .ffull_o              (txfifo_full_w), 
    .fempty_o             (ftempty_w), 
    .data_o               (txfifo_datao_w),
    .fifo_count_o         ()
  );
  
  fifo rx_fifo_top (
    .clk_i                (clk_i), 
    .rst_i                (rst_i), 
    .we_i                 (rxfifo_wri_w), 
    .rd_i                 (rdrx_neg), 
    .data_i               (rxfifo_datai_w),
    .ffull_o              (frfull_w), 
    .fempty_o             (rxfifo_empty_w), 
    .data_o               (rxfifo_datao_w),
    .fifo_count_o         (rxfifo_count_w)  
  );
  
  tx_cntr tx_cntr_top (
    .clk_i                (clk_i), 
    .baud_clk_i           (baudout_w), 
    .rst_i                (rst_i), 
    .ftempty_i            (ftempty_w), 
    .data_i               (txfifo_datao_w),
    .sout_o               (sout_o), 
    .rd_o                 (txfifo_rdi_w)
  );
  
  rx_cntr rx_cntr_top (
    .clk_i                (clk_i), 
    .baud_clk_i           (baudout_w), 
    .rst_i                (rst_i), 
    .ffull_i              (frfull_w), 
    .sin_i                (sin_i),
    .data_o               (rxfifo_datai_w), 
    .rx_overrun_o         (rx_overrun_w), 
    .wr_o                 (rxfifo_wri_w), 
    .frame_err_o          (rxframe_err_w)
  );
  
  always @ (posedge clk_i)
    begin
      rd_avail_pos <= rd_avail_neg;
    end

   
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          fifo_th <= 8'h0;
          line_control <= 8'h0;
          divisor_latch_l <= 8'h10;        // Baud rate 38400 bps (value 8'h08@10MHz)             
          divisor_latch_m <= 8'h0;  
          irq_rxcount <= 1'b0;
          irq_txfifo_empty <= 1'b0;
          irq_ident <= 8'h0;
          irq_enable <= 8'h0;
          txfifo_wei_r <= 1'b0;
        end
      else
        begin
          case ({adr_i,cs_i,wr_i})              // Synthesis full case
            5'b00011:
              begin
                if (line_control[7])
                  divisor_latch_l <= data_i;
                else if (~line_control[7])
                  txfifo_datai_r <= data_i;                
              end
            5'b00111:
              begin
                if (line_control[7])
                  divisor_latch_m <= data_i;
                else if (~line_control[7])
                  irq_enable <= data_i;
              end
            5'b01011: fifo_control <= data_i;
            5'b01111: line_control <= data_i;
            5'b10011: fifo_th <= data_i;
          endcase
          if ({adr_i,cs_i,wr_i,line_control[7]} === 6'b000110)
            txfifo_wei_r <= 1'b1;
          else
            txfifo_wei_r <= 1'b0;
            
          if (rxfifo_count_w > {1'b0,fifo_th})
            begin
              irq_rxcount <= 1'b1;
              irq_ident[0] <= 1'b1;
            end
          else
            begin
              irq_rxcount <= 1'b0;
              irq_ident[0] <= 1'b0;
            end
          if (ftempty_w)
            begin
              irq_txfifo_empty <= 1'b1;
              irq_ident[1] <= 1'b1;
            end
          else
            begin
              irq_txfifo_empty <= 1'b0;
              irq_ident[1] <= 1'b0;
            end
          if (rx_overrun_w)
            irq_ident[2] <= 1'b1;
          else
            irq_ident[2] <= 1'b0;
            
          if (rxframe_err_w)
            irq_ident[3] <= 1'b1;
          else
            irq_ident[3] <= 1'b0;
        end
      rx_overrun_neg <= rx_overrun_w;
      rxframe_err_neg <= rxframe_err_w; 
      rdrx_neg <= rd_rx;	   
      rd_avail_neg <= (rd_i & cs_i);
    end
    
endmodule
